# -*- coding: utf-8 -*-

from __future__ import absolute_import

import collections
import datetime
import json
import os.path
import sys
from .configuration import CONFIG_DIRECTORY
from ..util.logger import Logger

logger = Logger("config.cutoff_history")
logger.setstream("default", sys.stdout, Logger.WARNING)


DEFAULT_CONFIG_FILE = os.path.expanduser(os.path.join(CONFIG_DIRECTORY, "cutoff_history.json"))


class HistoryEntry(collections.namedtuple("HistoryEntry", ["filename", "frame", "time", "radii"])):
    class Date(object):
        def __init__(self, *args, **kwargs):
            if not isinstance(args[0], datetime.datetime):
                self._date = datetime.datetime(*args, **kwargs)
            else:
                datetime_obj = args[0]
                self._date = datetime_obj

        def __str__(self):
            return repr(self)

        def __repr__(self):
            return self._date.strftime("%m/%d/%y, %H:%M")

        @property
        def datetime_obj(self):
            return self._date

    def __new__(cls, filename, frame, time, radii):
        if isinstance(time, datetime.datetime):
            time = HistoryEntry.Date(time)
        else:
            try:
                time = HistoryEntry.Date(datetime.datetime.strptime(time, "%m/%d/%y, %H:%M"))
            except ValueError:
                try:
                    time = HistoryEntry.Date(datetime.datetime.strptime(time, "%Y-%m-%dT%H:%M:%S"))
                except ValueError:
                    time = HistoryEntry.Date(datetime.datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%f"))
        self = super(HistoryEntry, cls).__new__(cls, filename, frame, time, radii)
        return self


class CutoffHistory(object):
    def __init__(self, config_filepath=DEFAULT_CONFIG_FILE):
        self._config_filepath = config_filepath
        self._history = None
        self.read()

    def read(self):
        try:
            with open(self._config_filepath, "r") as f:
                history_json = json.load(f)
            #  for elem in history_json:
            #  elem[3] = {elem.encode('utf-8'): value for elem, value in elem[3]}
            self._history = [HistoryEntry(*entry) for entry in history_json]
        except IOError:
            self._history = []

    def save(self):
        def serialization_helper(obj):
            if isinstance(obj, HistoryEntry.Date):
                return obj.datetime_obj.isoformat()
            else:
                print(type(obj))
                raise TypeError

        try:
            history = []
            for entry in self.history:
                radii = {}
                for elem in entry[3]:
                    orig = elem
                    if isinstance(elem, bytes):
                        elem = elem.decode("utf-8")
                    radii[elem] = entry[3][orig]
                history.append(HistoryEntry(entry[0], entry[1], str(entry[2]), radii))
            with open(self._config_filepath, "w") as f:
                json.dump(history, f, default=serialization_helper)
        except IOError:
            logger.warn("Could not save cutoff radii history")
            raise

    def filtered_history(
        self,
        elements,
        entries_with_additional_elements=True,
        preferred_filenames_with_frames=None,
    ):
        elements = set(elements)
        filtered_history = []
        for entry in self._history:
            current_elements = set(entry.radii.keys())
            if (entries_with_additional_elements and elements.issubset(current_elements)) or (
                not entries_with_additional_elements and elements == current_elements
            ):
                filtered_history.append(entry)
        if preferred_filenames_with_frames is not None:

            def key_func(entry):
                """Key function thats for sorting by filenames. It preferres filenames from a given list by using
                integer weights.
                """
                if entry.filename in preferred_filenames_with_frames:
                    if entry.frame in preferred_filenames_with_frames[entry.filename]:
                        weight = 0
                    else:
                        weight = 1
                else:
                    weight = 2
                return (weight, entry.filename)

            filtered_history.sort(key=key_func)
        return filtered_history

    @property
    def history(self):
        return list(self._history)

    def add(self, entry):
        self._history.insert(0, entry)

    def extend(self, entry_list):
        entry_list = sorted(entry_list, key=lambda x: x.time, reverse=True)
        for entry in entry_list:
            self.add(entry)

    def remove(self, entry):
        self._history.remove(entry)

    def remove_list(self, entry_list):
        for entry in entry_list:
            self.remove(entry)


cutoff_history = CutoffHistory()



# -*- coding: utf-8 -*-

import os
import sys
from pymoldyn.util import env_is_true
from pymoldyn.util import message
from pymoldyn.util.logger import Logger

__all__ = [
    "atomstogrid",
    "mark_cavities",
    "cavity_triangles",
    "cavity_intersections",
    "mark_translation_vectors",
]

logger = Logger("core.calculation")
logger.setstream("default", sys.stdout, Logger.WARNING)


class CExtensionError(Exception):
    pass


try:
    from .extension_ctypes import (
        atomstogrid,
        mark_cavities,
        cavity_triangles,
        cavity_intersections,
        mark_translation_vectors,
    )
except OSError as e:
    if env_is_true("PYMOLDYN_FORCE_EXTENSIONS"):
        logger.error(e.__repr__())
        logger.error("C extension could not be loaded")
        message.log("C extension could not be loaded, aborting!")
        raise CExtensionError('"algorithm" C extension could not be loaded') from e
    else:
        logger.warn(e.__repr__())
        logger.warn("Falling back to Python functions")
        message.log(
            "C extensions could not be loaded, falling back to Python functions. Calculations may be very slow!"
        )
        from .extension_python import (
            atomstogrid,
            mark_cavities,
            cavity_triangles,
            cavity_intersections,
            mark_translation_vectors,
        )



import itertools

from .domain_centers import calculate_domain_centers as calc_dom
from .util.pos_bool_type import PosBoolType
from .util.numpy_extension import find_index_of_first_element_not_equivalent
from .util.node_border_iterator import iterate_node_border_with_adjacent_node_cells
import numpy as np

it = itertools.count(0, 1)


class ObjectType:
    DOMAIN = next(it)
    CAVITY = next(it)


def is_homogenous_split(data_part, mask_part):
    try:
        return PosBoolType(
            find_index_of_first_element_not_equivalent.find_index_of_first_element_not_equivalent(data_part, mask_part)
        )
    except TypeError:  # TODO fix logging
        if data_part[0][0][0] == 0:
            if mask_part[0][0][0] == 0:
                ret = np.where(np.logical_or(mask_part != 0, data_part != 0), 1, 0).nonzero()  # In c it is or not and
            else:
                ret = np.where(np.logical_or(mask_part == 0, data_part != 0), 1, 0).nonzero()
        else:
            if mask_part[0][0][0] == 0:
                ret = np.where(np.logical_or(mask_part != 0, data_part == 0), 1, 0).nonzero()
            else:
                ret = np.where(np.logical_or(mask_part == 0, data_part == 0), 1, 0).nonzero()
        try:
            return PosBoolType((ret[0][0], ret[1][0], ret[2][0]))
        except:
            return PosBoolType((-1, -1, -1))


def is_homogenous_merge(image_data_part, image_merge_data):
    def sign(x):
        return 1 if x > 0 else -1 if x < 0 else 0

    return sign(image_data_part[0, 0, 0]) == sign(image_merge_data[0, 0, 0])


def is_relevant_part(hom_image_data_part, object_type):
    obj_to_func = {ObjectType.DOMAIN: is_domain_part, ObjectType.CAVITY: is_cavity_part}
    return obj_to_func[object_type](hom_image_data_part)


def is_domain_part(hom_image_data_part):
    return hom_image_data_part[0, 0, 0] == 0


def is_cavity_part(hom_image_data_part):
    return hom_image_data_part[0, 0, 0] < 0


def is_inside_volume(hom_mask_part):
    return not bool(hom_mask_part[0, 0, 0])


def is_neighboring(node1, node2):
    x_, y_, z_ = node1[0]
    w_, h_, d_ = node1[1]
    x, y, z = node2[0]
    w, h, d = node2[1]
    return x - w_ <= x_ <= x + w and y - h_ <= y_ <= y + h and z - d_ <= z_ <= z + d


def split(data, mask, graph, object_type):
    node = ((0, 0, 0), data.shape)
    graph.set_initial_node(node)
    stack = [node]
    while len(stack) > 0:
        pos, dim = stack.pop()
        is_hom = is_homogenous_split(
            data[
                pos[0] : pos[0] + dim[0],
                pos[1] : pos[1] + dim[1],
                pos[2] : pos[2] + dim[2],
            ],
            mask[
                pos[0] : pos[0] + dim[0],
                pos[1] : pos[1] + dim[1],
                pos[2] : pos[2] + dim[2],
            ],
        )
        if not is_hom:
            nodes = graph.split_node((pos, dim), is_hom)
            for node in nodes:
                stack.append(node)
        elif is_relevant_part(
            data[
                pos[0] : pos[0] + dim[0],
                pos[1] : pos[1] + dim[1],
                pos[2] : pos[2] + dim[2],
            ],
            object_type,
        ) or not is_inside_volume(
            mask[
                pos[0] : pos[0] + dim[0],
                pos[1] : pos[1] + dim[1],
                pos[2] : pos[2] + dim[2],
            ]
        ):
            graph.remove_node((pos, dim))
    graph.forbid_splitting()


def merge(data, graph):
    for node, neighbors in graph.items():
        for neighbor in neighbors:
            if not graph.is_merged(node, neighbor):
                pos_node, dim_node = node
                pos_neighbor, dim_neighbor = neighbor
                data_node = data[
                    pos_node[0] : pos_node[0] + dim_node[0],
                    pos_node[1] : pos_node[1] + dim_node[1],
                    pos_node[2] : pos_node[2] + dim_node[2],
                ]
                data_neighbor = data[
                    pos_neighbor[0] : pos_neighbor[0] + dim_neighbor[0],
                    pos_neighbor[1] : pos_neighbor[1] + dim_neighbor[1],
                    pos_neighbor[2] : pos_neighbor[2] + dim_neighbor[2],
                ]
                if is_homogenous_merge(data_node, data_neighbor):
                    graph.merge_nodes(node, neighbor)


def add_periodic_neighbors(graph):
    border_node_translation_vectors = graph.get_border_node_translation_vectors()
    border_nodes = list(border_node_translation_vectors.keys())
    for i, n in enumerate(border_nodes[:-1]):
        for m in border_nodes[i + 1 :]:
            m_x, m_y, m_z = m[0]
            for translation_vector in border_node_translation_vectors[m]:
                translated_node = (
                    (
                        m_x + translation_vector[0],
                        m_y + translation_vector[1],
                        m_z + translation_vector[2],
                    ),
                    m[1],
                )
                if is_neighboring(n, translated_node):
                    graph.add_neighbors(n, [m], translation_vectors=[translation_vector])
                    break


def merge_periodic_border(data, graph):
    for border_node, border_neighbors in graph.iter_border_items():
        for border_neighbor in border_neighbors:
            if not graph.is_merged(border_node, border_neighbor, detect_cyclic_merge=True):
                pos_node, dim_node = border_node
                pos_neighbor, dim_neighbor = border_neighbor
                data_node = data[
                    pos_node[0] : pos_node[0] + dim_node[0],
                    pos_node[1] : pos_node[1] + dim_node[1],
                    pos_node[2] : pos_node[2] + dim_node[2],
                ]
                data_neighbor = data[
                    pos_neighbor[0] : pos_neighbor[0] + dim_neighbor[0],
                    pos_neighbor[1] : pos_neighbor[1] + dim_neighbor[1],
                    pos_neighbor[2] : pos_neighbor[2] + dim_neighbor[2],
                ]
                if is_homogenous_merge(data_node, data_neighbor):
                    graph.merge_nodes(border_node, border_neighbor)


def mark_domain_points(data, areas):
    for i, area in enumerate(areas):
        for node in area:
            x, y, z = node[0]
            w, h, d = node[1]
            data[x : x + w, y : y + h, z : z + d] = -(i + 1)


def calculate_domain_centers(atoms, combined_translation_vectors, areas):
    combined_translation_vectors_tuples = [tuple(t) for t in combined_translation_vectors]
    areas = [list(a) for a in areas]
    return calc_dom.calculate_domain_centers(atoms, combined_translation_vectors_tuples, areas)


def get_domain_area_cells(areas):
    cell_positions = [
        [
            (pos[0] + x, pos[1] + y, pos[2] + z)
            for pos, dim in area
            for x, y, z in itertools.product(*(range(c) for c in dim))
        ]
        for area in areas
    ]
    return cell_positions


def get_domain_surface_cells(data, mask, areas):
    domain = None

    def func(border_x, border_y, border_z, adjacent_node_cells):
        if bool(data[border_x, border_y, border_z]) or bool(mask[border_x, border_y, border_z]):
            for n in adjacent_node_cells:
                domain.add(n)

    domains = []
    for area in areas:
        domain = set()
        for node in area:
            iterate_node_border_with_adjacent_node_cells(node, func)
        domains.append(list(domain))
    return domains



# -*- coding: utf-8 -*-


# TODO: fix frame indexing. it should start at 0 for everything, except in the strings that are displayed in the widgets


from PySide6 import QtCore, QtGui, QtWidgets


class LabeledFrameChooser(QtWidgets.QWidget):
    value_changed = QtCore.Signal()

    def __init__(self, parent, num_frames, calculated, text):
        QtWidgets.QWidget.__init__(self, parent)

        self.framebar = FrameBar(self, num_frames, calculated)
        self.text = text
        self.num_frames = num_frames

        self.init_gui()

    def init_gui(self):

        hbox = QtWidgets.QHBoxLayout()
        vbox = QtWidgets.QVBoxLayout()

        self.lineedit = QtWidgets.QLineEdit(self)

        #        self.lineedit.setMinimumSize(30, 1)
        #        self.lineedit.setMaximumSize(40, 40)

        self.update_lineedit()
        self.lineedit.setAlignment(QtCore.Qt.AlignRight)
        self.lineedit.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        hbox.addWidget(QtWidgets.QLabel(self.text + ":", self), 0)
        hbox.addWidget(self.lineedit, 1)
        # hbox.addWidget(QtWidgets.QLabel('/'+str(self.maxf), self))
        # hbox.addStretch()

        vbox.addLayout(hbox)
        vbox.addWidget(self.framebar)

        self.lineedit.returnPressed.connect(self.lineedit_return_pressed)

        self.setLayout(vbox)
        self.show()

    def emit_value_changed(self):
        self.value_changed.emit()

    #    def load_dataset(self, filename):
    #        self.framebar.load_dataset(filename)

    def set_calculated_frames(self, calculated):
        self.framebar.set_calculated(calculated)

    def lineedit_return_pressed(self):
        try:
            l_1 = [int(i.strip()) for i in str(self.lineedit.text()).split(",")]
            # translate indices from human to machine
            l = [i - 1 for i in l_1]
        except ValueError:
            print("Enter a valid number")
        self.framebar.set_selection(l)

    def mousePressEvent(self, e):
        self.update_lineedit()

    def mouseMoveEvent(self, e):
        self.update_lineedit()

    def update_lineedit(self):
        sel = self.framebar.get_selection()
        # translate indices from machine to human
        sel_1 = [i + 1 for i in sel]
        s_1 = ", ".join([str(i) for i in sel_1])
        self.lineedit.setText(s_1)

    def value(self):
        return self.framebar.get_selection()


class FrameBar(QtWidgets.QWidget):

    def __init__(self, parent, num_frames, calculated):
        QtWidgets.QWidget.__init__(self, parent)

        self.parent = parent
        self.num_frames = num_frames
        self.calculated = calculated
        self.width = 300
        self.height = 10
        self.selection = [0]
        self.last_clicked = 0

        self.setMinimumSize(self.width + 5, self.height + 5)

    def draw(self):
        red = QtGui.QColor(255, 180, 180)
        green = QtGui.QColor(180, 255, 180)
        blue = QtGui.QColor(180, 180, 255)
        self.h = float(self.width) / (self.num_frames)

        p = self.painter

        p.setPen(QtCore.Qt.NoPen)

        p.setBrush(red)
        p.drawRect(0, 0, self.width, self.height)
        p.setBrush(green)
        for frame in self.calculated:
            i = frame
            p.drawRect(i * self.h, 1, self.h, self.height - 1)
        p.setBrush(blue)
        for sel in self.selection:
            i = sel
            p.drawRect(i * self.h, 1, self.h, self.height - 1)
        p.setPen(QtCore.Qt.SolidLine)
        p.setBrush(QtCore.Qt.NoBrush)
        for i in range(self.num_frames):
            p.drawLine((i + 1) * self.h, 0, (i + 1) * self.h, self.height)
        p.drawRect(0, 0, self.width, self.height)

    def mousePressEvent(self, e):
        self.process_mouse_press(e)
        e.ignore()

    def set_calculated(self, calculated):
        self.calculated = calculated
        self.repaint()

    def process_mouse_press(self, e):
        if e.buttons() and QtCore.Qt.LeftButton:
            clicked = int(e.x() / self.h)
            if 0 < e.x() < self.width and 0 < e.y() < self.height:
                self.process_mouse_event(clicked, e)
        e.ignore()

    def process_mouse_event(self, clicked, e):
        if e.modifiers() == QtCore.Qt.ShiftModifier:
            if self.last_clicked:
                for i in range(min(self.last_clicked, clicked), max(self.last_clicked, clicked) + 1):
                    if i not in self.selection:
                        self.selection.append(i)
                self.selection.sort()
        elif e.modifiers() == QtCore.Qt.CTRL:
            if clicked not in self.selection:
                self.selection.append(clicked)
                self.selection.sort()
        else:
            self.selection = [clicked]
        self.last_clicked = clicked
        self.repaint()
        self.parent.emit_value_changed()

    def mouseMoveEvent(self, e):
        if e.buttons() and QtCore.Qt.LeftButton:
            clicked = int(e.x() / self.h)
            if 0 < e.x() < self.width:
                self.process_mouse_event(clicked, e)
        e.ignore()

    def get_selection(self):
        return self.selection

    def set_selection(self, l):
        for value in l:
            if not 0 <= value < self.num_frames:
                l.remove(value)
            self.selection = l
            self.repaint()

    def paintEvent(self, e):
        self.painter = QtGui.QPainter()
        self.painter.begin(self)
        self.draw()
        self.painter.end()


#    def load_dataset(self, filename):
#        n_frames = calculation.count_frames(filename)
#        self.minf = 1
#        self.maxf = n_frames
#        self.calculated = [0]
#        self.selection = [0]
#        self.repaint()



import jinja2
import numpy as np
from PySide6 import QtCore, QtWidgets
import os.path
from ....core import elements
from ....core import bonds
from collections import Counter
from ....core.calculation.discretization import Discretization
from ..statistics.tree_list import TreeList
from ...util.webview import WebWidget


def render_html_atom_group(atom_number, atom_elements):
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "atoms.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = {
        "title": "Summary of atoms",
        "description": "a summary of all calculated atoms",
        "atom_number": atom_number,
        "atom_elements": atom_elements,
    }

    return template.render(template_vars)


def render_html_atom(
    index,
    atom_fullname,
    atom_positions,
    atom_number,
    covalent_radius,
    cutoff_radius,
    atom_color_rgb,
    bonds,
):
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "atom.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = {
        "title": "Summary of atoms",
        "description": "a summary of one atom",
        "index": index,
        "atom_fullname": atom_fullname,
        "atom_positions": atom_positions,
        "atom_number": atom_number,
        "covalent_radius": covalent_radius,
        "cutoff_radius": cutoff_radius,
        "atom_color_rgb": atom_color_rgb,
        "bonds": bonds,
    }

    return template.render(template_vars)


def render_html_cavity_center_group(number, surface_area, surface_volumes, volume_fraction):
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "cavities_center.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = {
        "title": "Summary of all cavities",
        "description": "a summary of all calculated center bases cavities",
        "number": number,
        "surface_area": surface_area,
        "surface_volumes": surface_volumes,
        "volume_fraction": volume_fraction,
    }

    return template.render(template_vars)


def render_html_cavity_center_group_unknown():
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "cavities_center_unknown.html"
    template = template_env.get_template(TEMPLATE_FILE)
    return template.render({})


def render_html_cavity_center(**kwargs):
    needed_keys = (
        "index",
        "surface",
        "volume",
        "domains",
        "volume_fraction",
        "mass_center",
        "squared_gyration_radius",
        "asphericity",
        "acylindricity",
        "anisotropy",
        "characteristic_radius",
        "is_cyclic",
    )
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "cavity_center.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = dict(kwargs)
    template_vars["title"] = "Summary of one cavity (domain)"
    template_vars["description"] = "a summary of one calculated cavities (domain)"
    if kwargs["surface"] is not None and kwargs["volume"] is not None:
        template_vars["surface_to_volume_ratio"] = kwargs["surface"] / kwargs["volume"]
    missing_values = tuple(key for key in needed_keys if key not in kwargs or kwargs[key] is None)
    template_vars.update((key, None) for key in missing_values)
    template_vars["missing_values"] = missing_values if len(missing_values) > 0 else None

    return template.render(template_vars)


def render_html_cavity_surface_group(number, surface_area, surface_volumes, volume_fraction):
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "cavities_surface.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = {
        "title": "Summary of atoms",
        "description": "a summary of all calculated surface based cavities",
        "number": number,
        "surface_area": surface_area,
        "surface_volumes": surface_volumes,
        "volume_fraction": volume_fraction,
    }

    return template.render(template_vars)


def render_html_cavity_surface_group_unknown():
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "cavities_surface_unknown.html"
    template = template_env.get_template(TEMPLATE_FILE)
    return template.render({})


def render_html_cavity_surface(**kwargs):
    needed_keys = (
        "index",
        "surface",
        "volume",
        "domains",
        "volume_fraction",
        "mass_center",
        "squared_gyration_radius",
        "asphericity",
        "acylindricity",
        "anisotropy",
        "characteristic_radius",
        "is_cyclic",
    )
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "cavity_surface.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = dict(kwargs)
    template_vars["title"] = "Summary of one cavity (domain)"
    template_vars["description"] = "a summary of one calculated cavities (domain)"
    if kwargs["surface"] is not None and kwargs["volume"] is not None:
        template_vars["surface_to_volume_ratio"] = kwargs["surface"] / kwargs["volume"]
    missing_values = tuple(key for key in needed_keys if key not in kwargs or kwargs[key] is None)
    template_vars.update((key, None) for key in missing_values)
    template_vars["missing_values"] = missing_values if len(missing_values) > 0 else None

    return template.render(template_vars)


def render_html_cavity_domain_group(number, surface_area, surface_volumes, surface_volumes_fractions):
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "domains.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = {
        "title": "Summary of all cavities (domains)",
        "description": "a summary of all calculated cavities (domains)",
        "number": number,
        "surface_area": surface_area,
        "surface_volumes": surface_volumes,
        "surface_volumes_fractions": surface_volumes_fractions,
    }

    return template.render(template_vars)


def render_html_cavity_domain_group_unknown():
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "domains_unknown.html"
    template = template_env.get_template(TEMPLATE_FILE)
    return template.render({})


def render_html_cavity_domain(**kwargs):
    needed_keys = (
        "index",
        "center",
        "surface",
        "volume",
        "volume_fraction",
        "surface_cavity_index",
        "center_cavity_index",
        "mass_center",
        "squared_gyration_radius",
        "asphericity",
        "acylindricity",
        "anisotropy",
        "characteristic_radius",
        "is_cyclic",
    )
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.join(os.path.dirname(__file__), "templates"))
    template_env = jinja2.Environment(loader=template_loader)

    TEMPLATE_FILE = "domain.html"
    template = template_env.get_template(TEMPLATE_FILE)

    template_vars = dict(kwargs)
    template_vars["title"] = "Summary of one cavity (domain)"
    template_vars["description"] = "a summary of one calculated cavities (domain)"
    if kwargs["surface"] is not None and kwargs["volume"] is not None:
        template_vars["surface_to_volume_ratio"] = kwargs["surface"] / kwargs["volume"]
    missing_values = tuple(key for key in needed_keys if key not in kwargs or kwargs[key] is None)
    template_vars.update((key, None) for key in missing_values)
    template_vars["missing_values"] = missing_values if len(missing_values) > 0 else None

    return template.render(template_vars)


class HTMLWindow(QtWidgets.QWidget):

    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        box = QtWidgets.QVBoxLayout()
        self.webview = WebWidget(css_filepath="../tabs/statistics/templates/style.css")

        self.atoms = None
        self.cavities_center = None
        self.cavities_surface = None
        self.domains = None
        self.discretization = None

        self.tree_list = None
        self.webview.set_gui_html(None)
        self.webview.gui_link_clicked.connect(self.link_clicked)

        box.addWidget(self.webview)
        self.setLayout(box)
        box.setContentsMargins(5, 0, 0, 0)
        self.show()

    def minimumSizeHint(self):
        return QtCore.QSize(150, -1)

    def sizeHint(self):
        return QtCore.QSize(250, -1)

    def link_clicked(self, value):
        """
        examines the data of the given link by *data* an calls the specific method to render the new HTML page
        :param data: Value of the link clicked in Webview
        :return: None
        """
        value = value.split("/")
        if value[0] == "surface_cavity":
            self.show_surface_cavity(int(value[1]) - 1)
        elif value[0] == "center_cavity":
            self.show_center_cavity(int(value[1]) - 1)
        elif value[0] == "domain":
            self.show_domain(int(value[1]) - 1)
        elif value[0] == "focus":
            position = [float(value[1]), float(value[2]), float(value[3])]
            self.window().control.visualization.set_focus_on(*position)
            self.window().center.gl_widget.update()
            self.window().center.combo.setCurrentIndex(0)
            self.window().center.gl_stack.setCurrentIndex(0)
        elif value[0] == "atom":
            self.show_atom(int(value[1]))
        elif value[0] == "hideothers":
            parent = self.parent()
            while parent.parent():
                parent = parent.parent()
            main_window = parent
            view_tab = main_window.view_dock.view_tab
            if value[1] == "atom":
                atom_index = int(value[2]) - 1
                view_tab.atom_check.indices = [atom_index]
                view_tab.atom_check.selection_checkbox_set_checked(True)
            elif value[1] == "element":
                element = elements.names[int(value[2])]
                visible_atom_indices = []
                for i, element_name in enumerate(self.atoms.elements):
                    if elements.names[elements.numbers[element_name.upper()]] == element:
                        visible_atom_indices.append(i)
                view_tab.atom_check.indices = visible_atom_indices
                view_tab.atom_check.selection_checkbox_set_checked(True)
            elif value[1] == "domain":
                domain_index = int(value[2]) - 1
                view_tab.domain_check.indices = [domain_index]
                view_tab.domain_check.selection_checkbox_set_checked(True)
            elif value[1] == "surface_cavity":
                surface_cavity_index = int(value[2]) - 1
                view_tab.surface_cavity_check.indices = [surface_cavity_index]
                view_tab.surface_cavity_check.selection_checkbox_set_checked(True)
            elif value[1] == "center_cavity":
                center_cavity_index = int(value[2]) - 1
                view_tab.center_cavity_check.indices = [center_cavity_index]
                view_tab.center_cavity_check.selection_checkbox_set_checked(True)
        elif value[0] == "addtovisible":
            parent = self.parent()
            while parent.parent():
                parent = parent.parent()
            main_window = parent
            view_tab = main_window.view_dock.view_tab
            if value[1] == "atom":
                atom_index = int(value[2]) - 1
                view_tab.domain_check.add_indices([atom_index])
            elif value[1] == "domain":
                domain_index = int(value[2]) - 1
                view_tab.domain_check.add_indices([domain_index])
            elif value[1] == "surface_cavity":
                surface_cavity_index = int(value[2]) - 1
                view_tab.surface_cavity_check.add_indices([surface_cavity_index])
            elif value[1] == "center_cavity":
                center_cavity_index = int(value[2]) - 1
                view_tab.center_cavity_check.add_indices([center_cavity_index])
        elif value[0] == "showall":
            parent = self.parent()
            while parent.parent():
                parent = parent.parent()
            main_window = parent
            view_tab = main_window.view_dock.view_tab
            if value[1] == "atoms":
                view_tab.atom_check.setChecked(True)
                view_tab.atom_check.selection_checkbox_set_checked(False)
            if value[1] == "domains":
                view_tab.domain_check.setChecked(True)
                view_tab.domain_check.selection_checkbox_set_checked(False)
            if value[1] == "surface_cavities":
                view_tab.surface_cavity_check.setChecked(True)
                view_tab.surface_cavity_check.selection_checkbox_set_checked(False)
            if value[1] == "center_cavities":
                view_tab.center_cavity_check.setChecked(True)
                view_tab.center_cavity_check.selection_checkbox_set_checked(False)
        elif value[0] == "recalculate":
            parent = self.parent()
            while parent.parent():
                parent = parent.parent()
            main_window = parent
            file_tab = main_window.file_dock.file_tab
            current_filename, current_frame = file_tab.last_shown_filename_with_frame
            file_tab.calculate({current_filename: [current_frame]})

    def update_results(self, results):
        self.atoms = results.atoms
        self.cavities_center = results.center_cavities
        self.cavities_surface = results.surface_cavities
        self.domains = results.domains
        self.discretization = Discretization(results.atoms.volume, results.resolution, True)
        self.show_atom_group()

    def show_atom_group(self):
        atom_number = self.atoms.number
        atom_elements = Counter(self.atoms.elements)

        self.webview.set_gui_html(render_html_atom_group(atom_number, atom_elements))

    def show_atom(self, index):
        if self.tree_list is not None:
            self.tree_list.select_atom(index)

        # for bond in bonds:
        #    if index not in self.atoms.bonds[bond]:
        #        self.atoms.bonds[bond].append(index)

        atom_name = self.atoms.elements[index]  # atom name from periodic systen
        atom_fullname = elements.names[elements.numbers[atom_name.decode("utf-8").upper()]]  # get full atom name
        atom_color_rgb = elements.colors[elements.numbers[atom_name.decode("utf-8").upper()]]
        atom_positions = self.atoms.positions[index]
        atom_number = elements.numbers[atom_name.decode("utf-8").upper()]
        covalent_radius = self.atoms.covalence_radii[index]
        cutoff_radius = self.atoms.radii[index]
        bonds = self.atoms.bonds[index]

        # print dir(self.domains[0])

        self.webview.set_gui_html(
            render_html_atom(
                index,
                atom_fullname,
                atom_positions,
                atom_number,
                covalent_radius,
                cutoff_radius,
                atom_color_rgb,
                bonds,
            )
        )

    def show_center_cavity_group(self):
        number = 0
        surface_area = 0.0
        volumes = 0.0
        volume_fraction = 0.0
        if self.cavities_center is None:
            self.webview.set_gui_html(render_html_cavity_center_group_unknown())
            return
        number = self.cavities_center.number
        for sf in self.cavities_center.surface_areas:
            surface_area += sf
        for vl in self.cavities_center.volumes:
            volumes += vl

        if self.atoms.volume is not None:
            volume_fraction = (volumes / self.atoms.volume.volume) * 100

        self.webview.set_gui_html(render_html_cavity_center_group(number, surface_area, volumes, volume_fraction))

    def show_center_cavity(self, index):
        if self.tree_list is not None:
            self.tree_list.select_center_cavity(index)
        attrs = self._create_attr_getter(self.cavities_center, index)
        data = {}
        data["index"] = index
        data["volume_fraction"] = 0.0
        cavities = attrs.multicavities
        domains = []
        for cavity in cavities:
            domains.append(
                (
                    cavity + 1,
                    self.discretization.discrete_to_continuous(self.domains.centers[cavity]),
                )
            )
        data["domains"] = domains

        data["surface"] = attrs.surface_areas
        data["volume"] = attrs.volumes
        data["mass_center"] = attrs.mass_centers
        data["squared_gyration_radius"] = attrs.squared_gyration_radii
        data["asphericity"] = attrs.asphericities
        data["acylindricity"] = attrs.acylindricities
        data["anisotropy"] = attrs.anisotropies
        data["characteristic_radius"] = attrs.characteristic_radii
        data["is_cyclic"] = index in self.cavities_center.cyclic_area_indices

        if self.atoms.volume is not None:
            data["volume_fraction"] = (data["volume"] / self.atoms.volume.volume) * 100

        self.webview.set_gui_html(render_html_cavity_center(**data))

    def show_surface_cavity_group(self):
        number = 0
        surface_area = 0.0
        volumes = 0.0
        volume_fraction = 0.0
        if self.cavities_surface is None:
            self.webview.set_gui_html(render_html_cavity_surface_group_unknown())
            return
        number = self.cavities_surface.number
        for sf in self.cavities_surface.surface_areas:
            surface_area += sf
        for vl in self.cavities_surface.volumes:
            volumes += vl

        if self.atoms.volume is not None:
            volume_fraction = (volumes / self.atoms.volume.volume) * 100

        self.webview.set_gui_html(render_html_cavity_surface_group(number, surface_area, volumes, volume_fraction))

    def show_surface_cavity(self, index):
        if self.tree_list is not None:
            self.tree_list.select_surface_cavity(index)
        attrs = self._create_attr_getter(self.cavities_surface, index)
        data = {}
        data["index"] = index
        data["volume_fraction"] = 0.0
        cavities = attrs.multicavities
        domains = []
        for cavity in cavities:
            domains.append(
                (
                    cavity + 1,
                    self.discretization.discrete_to_continuous(self.domains.centers[cavity]),
                )
            )
        data["domains"] = domains

        data["surface"] = attrs.surface_areas
        data["volume"] = attrs.volumes
        data["mass_center"] = attrs.mass_centers
        data["squared_gyration_radius"] = attrs.squared_gyration_radii
        data["asphericity"] = attrs.asphericities
        data["acylindricity"] = attrs.acylindricities
        data["anisotropy"] = attrs.anisotropies
        data["characteristic_radius"] = attrs.characteristic_radii
        data["is_cyclic"] = index in self.cavities_surface.cyclic_area_indices

        if self.atoms.volume is not None:
            data["volume_fraction"] = (data["volume"] / self.atoms.volume.volume) * 100

        self.webview.set_gui_html(render_html_cavity_surface(**data))

    def show_domain_group(self):
        number = 0
        surface = 0.0
        volumes = 0.0
        volume_fraction = 0.0

        if self.domains is None:
            self.webview.set_gui_html(render_html_cavity_domain_group_unknown())
            return
        number = self.domains.number
        for sf in self.domains.surface_areas:
            surface += sf
        for vl in self.domains.volumes:
            volumes += vl

        if self.atoms.volume is not None:
            volume_fraction = (volumes / self.atoms.volume.volume) * 100

        self.webview.set_gui_html(render_html_cavity_domain_group(number, surface, volumes, volume_fraction))

    def show_domain(self, index):
        if self.tree_list is not None:
            self.tree_list.select_domain(index)
        attrs = self._create_attr_getter(self.domains, index)
        data = {}
        data["index"] = index
        discrete_center = attrs.centers
        data["center"] = self.discretization.discrete_to_continuous(discrete_center)
        data["surface"] = attrs.surface_areas
        data["volume"] = attrs.volumes
        data["volume_fraction"] = 0.0
        data["surface_cavity_index"] = None
        data["center_cavity_index"] = None
        data["mass_center"] = attrs.mass_centers
        data["squared_gyration_radius"] = attrs.squared_gyration_radii
        data["asphericity"] = attrs.asphericities
        data["acylindricity"] = attrs.acylindricities
        data["anisotropy"] = attrs.anisotropies
        data["characteristic_radius"] = attrs.characteristic_radii
        data["is_cyclic"] = index in self.domains.cyclic_area_indices

        if self.atoms.volume is not None:
            data["volume_fraction"] = (data["volume"] / self.atoms.volume.volume) * 100
        if self.domains is not None and self.cavities_surface is not None:
            for i in range(len(self.cavities_surface.multicavities)):
                if index in self.cavities_surface.multicavities[i]:
                    data["surface_cavity_index"] = i + 1
                    break
        if self.domains is not None and self.cavities_center is not None:
            for i in range(len(self.cavities_center.multicavities)):
                if index in self.cavities_center.multicavities[i]:
                    data["center_cavity_index"] = i + 1
                    break

        self.webview.set_gui_html(render_html_cavity_domain(**data))

    @staticmethod
    def _create_attr_getter(obj, index):
        class AttrGetter:
            def __init__(self, obj, index):
                self._obj = obj
                self._index = index

            def __getattr__(self, attr):
                value = getattr(self._obj, attr)
                is_numpy_array = isinstance(value, np.ndarray)
                if (is_numpy_array and len(value.shape) > 0) or (not is_numpy_array and len(value) > 0):
                    return value[self._index]
                else:
                    return None

        return AttrGetter(obj, index)



#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import numpy as np
import numpy.linalg as la
import zlib
import struct


def cartesian(arrays, out=None):
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m : (j + 1) * m, 1:] = out[0:m, 1:]
    return out


def to_png(image):
    """
    Return a binary png image for a given image stored as two-dimensional numpy
    array
    """
    width, height, _ = image.shape
    image.shape = np.prod(image.shape)
    width_byte_4 = width * 4
    data_iterator = (
        b"\x00" + image[span : span + width_byte_4].tostring()
        for span in range((height - 1) * width * 4, -1, -width_byte_4)
    )
    raw_data = bytes().join(data_iterator)

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return struct.pack(b"!I", len(data)) + chunk_head + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))

    return b"".join(
        [
            b"\x89PNG\r\n\x1a\n",
            png_pack(b"IHDR", struct.pack(b"!2I5B", width, height, 8, 6, 0, 0, 0)),
            png_pack(b"IDAT", zlib.compress(raw_data, 9)),
            png_pack(b"IEND", b""),
        ]
    )


def write_png(image, filename):
    """Write an image stored as two-dimensional numpy array to a file"""
    with open(filename, "wb") as out:
        out.write(to_png(image))


def intensities_to_image(intensities):
    """
    Return a 32bit RGBA image representing the given intensities
    """
    width, height = intensities.shape
    image = np.zeros((width, height, 4), dtype=np.uint8)
    image[:, :, 2] = image[:, :, 1] = image[:, :, 0] = normalized(intensities.T) * 255
    image[:, :, 3] = 255
    return image


def intensities_to_image_unnormalized(intensities):
    """
    Return a 32bit RGBA image representing the given intensities
    """
    width, height = intensities.shape
    image = np.zeros((width, height, 4), dtype=np.uint8)
    image[:, :, 2] = image[:, :, 1] = image[:, :, 0] = intensities.T * 255
    image[:, :, 3] = 255
    return image


def normalized(array):
    """
    Return a normalized version of a numpy array, with values ranging from
    0 to 1
    """
    ptp = np.ptp(array)
    if ptp == 0:
        ptp = 1
    return (array - np.min(array)) / ptp


def create_rotation_matrix(angle, x, y, z):
    x, y, z = np.array((x, y, z)) / la.norm((x, y, z))
    matrix = np.zeros((3, 3), dtype=np.float32)
    cos = np.cos(angle)
    sin = np.sin(angle)
    matrix[0, 0] = x * x * (1 - cos) + cos
    matrix[1, 0] = x * y * (1 - cos) + sin * z
    matrix[0, 1] = x * y * (1 - cos) - sin * z
    matrix[2, 0] = x * z * (1 - cos) - sin * y
    matrix[0, 2] = x * z * (1 - cos) + sin * y
    matrix[1, 1] = y * y * (1 - cos) + cos
    matrix[1, 2] = y * z * (1 - cos) - sin * x
    matrix[2, 1] = y * z * (1 - cos) + sin * x
    matrix[2, 2] = z * z * (1 - cos) + cos
    return matrix


def create_perspective_projection_matrix(
    vertical_field_of_view,
    aspect_ratio,
    near_clipping_pane_distance,
    far_clipping_pane_distance,
):
    """
    Creates a perspective projection matrix as described at:
        http://www.opengl.org/sdk/docs/man2/xhtml/gluPerspective.xml
    """
    matrix = np.zeros((4, 4), dtype=np.float32)
    f = 1 / np.tan(0.5 * vertical_field_of_view)
    matrix[0, 0] = f / aspect_ratio
    matrix[1, 1] = f
    matrix[2, 2] = (near_clipping_pane_distance + far_clipping_pane_distance) / (
        near_clipping_pane_distance - far_clipping_pane_distance
    )
    matrix[2, 3] = (
        2
        * near_clipping_pane_distance
        * far_clipping_pane_distance
        / (near_clipping_pane_distance - far_clipping_pane_distance)
    )
    matrix[3, 2] = -1
    return matrix


def create_orthogonal_projection_matrix(left, right, bottom, top, near, far):
    matrix = np.zeros((4, 4), dtype=np.float32)
    matrix[0, 0] = 2 / (right - left)
    matrix[1, 1] = 2 / (top - bottom)
    matrix[2, 2] = 2 / (far - near)
    matrix[0, 3] = (right + left) / (right - left)
    matrix[1, 3] = (top + bottom) / (top - bottom)
    matrix[2, 3] = (far + near) / (far - near)
    matrix[3, 3] = 1
    return matrix


def create_rotation_matrix_homogenous(angle, x, y, z):
    matrix = np.eye(4, 4, dtype=np.float32)
    matrix[:3, :3] = create_rotation_matrix(angle, x, y, z)
    return matrix


def create_translation_matrix_homogenous(x, y, z):
    matrix = np.eye(4, 4, dtype=np.float32)
    matrix[:3, 3] = (x, y, z)
    return matrix


def normalize(x):
    return x / la.norm(x)


def create_look_at_matrix(eye, center, up):
    forward = normalize(center - eye)
    up = normalize(up)
    s = np.cross(forward, up)
    upward = np.cross(s, forward)

    matrix = np.eye(4, 4, dtype=np.float32)
    matrix[0, :3] = s
    matrix[1, :3] = upward
    matrix[2, :3] = -forward
    matrix[:3, 3] = -eye
    return matrix



#!/usr/bin/env python
# coding: utf-8

import os
import sys
from PIL import Image
import gr
import numpy as np


class GRVideoOutput(object):
    def __init__(self):
        self.is_recording = False

    def begin(self, outfile_name=None):
        self.is_recording = True
        os.environ["GKS_WSTYPE"] = "mov"
        if outfile_name is not None:
            os.environ["GKS_FILEPATH"] = os.path.abspath(outfile_name)

    def write(self, image, device_pixel_ratio=1):
        height, width = image.shape[:2]
        gr.clearws()
        if width > height:
            xmax = 1.0
            ymax = 1.0 * height / width
        else:
            xmax = 1.0 * width / height
            ymax = 1.0

        metric_width, metric_height, pixel_width, pixel_height = gr.inqdspsize()
        meter_per_horizontal_pixel = metric_width / pixel_width
        meter_per_vertical_pixel = metric_height / pixel_height
        gr.setwsviewport(
            0,
            meter_per_horizontal_pixel * width * device_pixel_ratio,
            0,
            meter_per_vertical_pixel * height * device_pixel_ratio,
        )
        gr.setwswindow(0, xmax, 0, ymax)
        gr.setviewport(0, xmax, 0, ymax)
        gr.setwindow(0, xmax, 0, ymax)
        gr.drawimage(
            0,
            xmax,
            0,
            ymax,
            width * device_pixel_ratio,
            height * device_pixel_ratio,
            image.view("uint32"),
        )
        gr.updatews()

    def end(self):
        self.is_recording = False
        gr.emergencyclosegks()


def main(argv):
    if len(argv) < 3:
        print("usage: python video_output.py outfile.mov frame1.png...")
        sys.exit(1)
    argv = list(map(os.path.abspath, argv[1:]))
    outfile_name = argv[0]
    frame_file_names = argv[1:]
    os.chdir(os.path.dirname(frame_file_names[0]))
    video_output = GRVideoOutput()
    video_output.begin(outfile_name)
    for frame_index, frame_file_name in enumerate(frame_file_names):
        image = np.array(Image.open(frame_file_name))
        video_output.write(image)
        sys.stdout.flush()
    video_output.end()


if __name__ == "__main__":
    main(sys.argv)



# -*- coding: utf-8 -*-


from .util.calc_table import CalculationTable, TableModel
from .cutoff_history_dialog import CutoffHistoryDialog
from ..cutoff_table_widget import CutoffTableWidget
from ..cutoff_preset_combobox import CutoffPresetComboBox
from ...core import calculation
from ...core import file
import collections
import collections.abc
import datetime
import os.path
import itertools as it
import re
from ...config.configuration import config
from ...config.cutoff_presets import cutoff_presets
from ...config.cutoff_presets import Preset
from ...config.cutoff_history import cutoff_history
from ...config.cutoff_history import HistoryEntry
from PySide6 import QtCore, QtWidgets


class CalculationSettingsDialog(QtWidgets.QDialog):

    RES_MIN = 32
    RES_MAX = 1024
    RES_INTERVAL = 32

    def __init__(self, parent, file_frame_dict):
        super().__init__(parent)

        self.control = parent.control
        self.resolution = config.Computation.std_resolution
        self.filenames = [f for f in file_frame_dict.keys()]
        self.file_frame_dict = file_frame_dict

        self.init_gui()
        self.setWindowTitle("Calculation Settings")

    def init_gui(self):

        vbox = QtWidgets.QVBoxLayout()
        hbox = QtWidgets.QHBoxLayout()
        inner_layout = QtWidgets.QVBoxLayout()
        button_hbox = QtWidgets.QHBoxLayout()
        res_hbox = QtWidgets.QHBoxLayout()

        self.res_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.res_slider.setMinimum(0)
        self.res_slider.setMaximum((self.RES_MAX - self.RES_MIN) / self.RES_INTERVAL)
        # self.res_slider.setTickInterval(1)
        # self.res_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.res_slider.valueChanged[int].connect(self.slider_changing)
        self.res_slider.sliderReleased.connect(self.slider_released)

        self.lineedit = QtWidgets.QLineEdit(self)
        self.lineedit.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.lineedit.setMinimumSize(30, 1)
        self.lineedit.setMaximumSize(40, 40)
        self.lineedit.returnPressed.connect(self.lineedit_return)

        res_hbox.addWidget(QtWidgets.QLabel("resolution:", self))
        res_hbox.addWidget(self.res_slider)
        res_hbox.addWidget(self.lineedit)

        # set font
        #        font = QFont("Courier New", 14)
        #        table_view.setFont(font)
        # set column width to fit contents (set font first!)
        # enable sortingq
        #        table_view.setSortingEnabled(True)

        ok_button = QtWidgets.QPushButton("Ok", self)
        ok_button.setAutoDefault(False)
        ok_button.clicked.connect(self.ok)

        cancel_button = QtWidgets.QPushButton("Cancel", self)
        cancel_button.setAutoDefault(False)
        cancel_button.clicked.connect(self.cancel)

        button_hbox.addStretch()
        button_hbox.addWidget(ok_button)
        button_hbox.addStretch()
        button_hbox.addWidget(cancel_button)
        button_hbox.addStretch()

        self.table_view = CalculationTable(self)
        self.res_slider.setValue((self.resolution - self.RES_MIN) / self.RES_INTERVAL)
        self.update_table()

        self.surf_check = QtWidgets.QCheckBox("calculate surface based cavities", self)
        self.surf_check.setChecked(True)
        self.center_check = QtWidgets.QCheckBox("calculate center based cavities", self)
        self.gyration_tensor_check = QtWidgets.QCheckBox("calculate gyration tensor parameters", self)
        self.gyration_tensor_check.setToolTip("squared_gyration_radius, asphericity, acylindricity, anisotropy")
        self.overwrite_check = QtWidgets.QCheckBox("overwrite existing results", self)
        self.exporthdf5_check = QtWidgets.QCheckBox("export results as HDF5 files", self)
        self.exporttext_check = QtWidgets.QCheckBox("export results as text files", self)
        self.exportdir_radio_input = QtWidgets.QRadioButton("export to the directory of the input files", self)
        self.exportdir_radio_config = QtWidgets.QRadioButton("export to %s" % config.Path.result_dir, self)

        self.exportdir_radio_input.setChecked(True)

        covalence_radii_by_element = self.__get_all_covalence_radii_by_element()
        self.radii_widget = RadiiWidget(covalence_radii_by_element, self.file_frame_dict, self)

        inner_layout.addLayout(res_hbox)
        inner_layout.addWidget(self.table_view)
        inner_layout.addWidget(self.surf_check)
        inner_layout.addWidget(self.center_check)
        inner_layout.addWidget(self.gyration_tensor_check)
        inner_layout.addWidget(self.overwrite_check)
        inner_layout.addWidget(self.exporthdf5_check)
        inner_layout.addWidget(self.exporttext_check)
        inner_layout.addWidget(self.exportdir_radio_input)
        inner_layout.addWidget(self.exportdir_radio_config)

        hbox.addLayout(inner_layout)
        hbox.addWidget(self.radii_widget)

        vbox.addLayout(hbox)
        vbox.addLayout(button_hbox)

        self.setLayout(vbox)

    def __get_all_covalence_radii_by_element(self):
        covalence_radii_by_element, elements_by_frame, element_combinations = self.__read_atom_info()
        return covalence_radii_by_element

    def __get_elements_by_frame(self):
        covalence_radii_by_element, elements_by_frame, element_combinations = self.__read_atom_info()
        return elements_by_frame

    def __get_element_combinations(self):
        covalence_radii_by_element, elements_by_frame, element_combinations = self.__read_atom_info()
        return element_combinations

    def __read_atom_info(self):
        method = self.__read_atom_info.__func__
        for attr in (
            "covalence_radii_by_element",
            "elements_by_frame",
            "element_combinations",
        ):
            if not hasattr(method, attr):
                setattr(method, attr, None)

        if (
            method.covalence_radii_by_element is None
            or method.elements_by_frame is None
            or method.element_combinations is None
        ):

            radii = {}
            elements_by_frame = {}
            element_combinations = set()
            for filepath, frames in self.file_frame_dict.items():
                elements_by_frame[filepath] = {}
                inputfile = file.File.open(filepath)
                if frames == (-1,):
                    frames = range(inputfile.info.num_frames)
                for frame in frames:
                    atoms = inputfile.getatoms(frame)
                    current_radii = atoms.covalence_radii_by_element
                    current_elements = atoms.elements
                    radii.update(current_radii)
                    elements_by_frame[filepath][frame] = current_elements
                    element_combinations.add(tuple(current_elements))
            method.covalence_radii_by_element = radii
            method.elements_by_frame = elements_by_frame
        return (
            method.covalence_radii_by_element,
            method.elements_by_frame,
            method.element_combinations,
        )

    def __update_cutoff_history(self):
        timestamp = datetime.datetime.now()
        user_cutoff_radii = self.radii_widget.cutoff_radii
        elements = self.__get_all_covalence_radii_by_element().keys()
        if not isinstance(user_cutoff_radii, collections.abc.Iterable):
            user_cutoff_radii = dict((elem, user_cutoff_radii) for elem in elements)
        elements_by_frame = self.__get_elements_by_frame()
        new_history = []
        for filepath, frames in self.file_frame_dict.items():
            filename = os.path.basename(filepath)
            file_elements = elements_by_frame[filepath]
            for frame in frames:
                elements = file_elements[frame]
                frame_cutoff_radii = dict((elem, user_cutoff_radii[elem]) for elem in elements)
                history_entry = HistoryEntry(filename, frame, timestamp, frame_cutoff_radii)
                new_history.append(history_entry)
        cutoff_history.extend(new_history)
        cutoff_history.save()

    def __update_cutoff_presets(self):
        save_preset_name = self.radii_widget.save_preset_name
        if save_preset_name is not None:
            user_cutoff_radii = self.radii_widget.cutoff_radii
            cutoff_presets.add(Preset(save_preset_name, user_cutoff_radii))
            cutoff_presets.save()

    def update_table(self):
        # get timestamps for selected frames for each file

        # surface based
        surface_ts = []
        center_ts = []
        for i, ts in enumerate(self.timestamps(center_based=False)):
            frames = (
                range(file.File.open(self.filenames[i]).info.num_frames)
                if self.file_frame_dict[self.filenames[i]][0] == -1
                else self.file_frame_dict[self.filenames[i]]
            )
            surface_ts.append([])
            for frame in frames:
                ts, temp = it.tee(ts)
                surface_ts[i].append(list(temp)[frame])

        # center based timestamps for the given frames
        center_ts = []
        for i, ts in enumerate(self.timestamps(center_based=True)):
            frames = (
                range(file.File.open(self.filenames[i]).info.num_frames)
                if self.file_frame_dict[self.filenames[i]][0] == -1
                else self.file_frame_dict[self.filenames[i]]
            )
            center_ts.append([])
            for frame in frames:
                ts, temp = it.tee(ts)
                center_ts[i].append(list(temp)[frame])

        # reduce to a single value per file
        surface_ts = ["X" if "X" in ts else ts[0] for ts in surface_ts]
        center_ts = ["X" if "X" in ts else ts[0] for ts in center_ts]
        basenames = [os.path.basename(path) for path in self.filenames]
        frames = [
            (
                str([frame + 1 for frame in self.file_frame_dict[f]])[1:-1]
                if not self.file_frame_dict[f][0] == -1
                else "all"
            )
            for f in self.filenames
        ]

        data_list = zip(basenames, surface_ts, center_ts, frames)

        # set table data
        header = ["dataset", "surface based", "center based", "frames"]
        table_model = TableModel(self, data_list, header)
        self.table_view.setModel(table_model)
        self.table_view.resizeColumnsToContents()

        # calculate table size to set its minimum size
        width = (self.table_view.model().columnCount(self.table_view) - 1) + self.table_view.verticalHeader().width()
        for i in range(self.table_view.model().columnCount(self.table_view)):
            width += self.table_view.columnWidth(i)
        self.table_view.setMinimumWidth(width)

        height = (self.table_view.model().rowCount(self.table_view) - 1) + self.table_view.horizontalHeader().height()
        for i in range(self.table_view.model().rowCount(self.table_view)):
            height += self.table_view.rowHeight(i)
        self.table_view.setMinimumHeight(height)

    def lineedit_return(self):
        try:
            self.resolution = int(self.lineedit.text())
            value = (self.resolution - self.RES_MIN) / self.RES_INTERVAL
            self.res_slider.setValue(value)
            self.update_table()
        except ValueError:
            pass

    def slider_changing(self, value):
        self.lineedit.setText(str(value * self.RES_INTERVAL + self.RES_MIN))

    def slider_released(self):
        self.resolution = self.res_slider.value() * self.RES_INTERVAL + self.RES_MIN
        self.update_table()

    def timestamps(self, center_based=False):
        return [
            self.control.calculation.calculatedframes(
                file.get_abspath(fn), self.resolution, not center_based, center_based
            ).prettystrings()
            for fn in self.filenames
        ]

    def ok(self):
        self.lineedit_return()
        self.__update_cutoff_history()
        self.__update_cutoff_presets()
        self.done(QtWidgets.QDialog.Accepted)

    def cancel(self):
        self.done(QtWidgets.QDialog.Rejected)

    def calculation_settings(self):
        calc_settings = None
        while True:
            ok = self.exec_()
            if ok != QtWidgets.QDialog.Accepted:
                break
            try:
                cutoff_radii = self.radii_widget.cutoff_radii
                surface_based = self.surf_check.isChecked()
                center_based = self.center_check.isChecked()
                gyration_tensor = self.gyration_tensor_check.isChecked()
                overwrite = self.overwrite_check.isChecked()
                exporthdf5 = self.exporthdf5_check.isChecked()
                exporttext = self.exporttext_check.isChecked()
                if self.exportdir_radio_config.isChecked():
                    exportdir = os.path.expanduser(config.Path.result_dir)
                else:
                    exportdir = None
                calc_settings = calculation.CalculationSettings(
                    datasets=self.file_frame_dict,
                    resolution=self.resolution,
                    cutoff_radii=cutoff_radii,
                    domains=True,
                    surface_cavities=surface_based,
                    center_cavities=center_based,
                    gyration_tensor=gyration_tensor,
                    recalculate=overwrite,
                    exporthdf5=exporthdf5,
                    exporttext=exporttext,
                    exportdir=exportdir,
                )
                break
            except ValueError:
                pass

        return (calc_settings, ok)


class RadiiWidget(QtWidgets.QWidget):
    class RadiiType:
        FIXED = 0
        CUSTOM = 1

    def __init__(self, radii, file_frame_dict, parent=None):
        super().__init__(parent)
        self._radii = radii
        self._file_frame_dict = file_frame_dict
        self._preferred_filenames_with_frames = dict(
            (os.path.basename(filepath), frames) for filepath, frames in self._file_frame_dict.items()
        )
        self._discard_preset_choice_on_next_rb_click = True  # Default value is True
        self._selected_radii_type = RadiiWidget.RadiiType.FIXED
        self._init_ui()
        self._init_cutoff_radii()

    def _init_ui(self):
        # Fixed Radio Button
        self.rb_fixed = QtWidgets.QRadioButton("Fixed Radius:")
        self.le_fixed = QtWidgets.QLineEdit()
        self.le_fixed.setMinimumWidth(150)
        self.le_fixed.setVisible(False)
        self.le_fixed.textEdited.connect(self.le_fixed_text_edited)
        self.rb_fixed.clicked.connect(self.rb_fixed_clicked)

        # QStackedWidget le_fixed
        self.tmp1 = QtWidgets.QWidget()
        self.sw_fixed = QtWidgets.QStackedWidget()
        self.sw_fixed.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.sw_fixed.addWidget(self.le_fixed)
        self.sw_fixed.addWidget(self.tmp1)
        self.sw_fixed.setCurrentIndex(1)

        # Custom Radio Button + Table
        self.rb_custom = QtWidgets.QRadioButton("Custom:")
        self.tw_cutoff = CutoffTableWidget(self._radii)
        self.tw_cutoff.text_edited.connect(self.tw_cutoff_text_edited)
        self.rb_custom.clicked.connect(self.rb_custom_clicked)

        # Preset Combo Box
        self.cb_preset = CutoffPresetComboBox()
        self.cb_preset.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.cb_preset.preset_selected.connect(self.cb_preset_selected)

        # History button
        self.pb_history = QtWidgets.QPushButton("History", self)
        self.pb_history.setMinimumWidth(0)
        self.pb_history.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.pb_history.setAutoDefault(False)
        self.pb_history.clicked.connect(self.pb_history_clicked)
        if (
            len(
                cutoff_history.filtered_history(
                    self._radii.keys(),
                    preferred_filenames_with_frames=self._preferred_filenames_with_frames,
                )
            )
            == 0
        ):
            self.pb_history.setVisible(False)

        # Preset save
        self.cb_preset_save = QtWidgets.QCheckBox("Save as Preset", self)
        self.le_preset_save = QtWidgets.QLineEdit()
        self.le_preset_save.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.le_preset_save.textChanged.connect(self.le_preset_save_changed)

        self.la_main = QtWidgets.QGridLayout()
        self.la_main.setContentsMargins(0, 0, 0, 0)
        self.la_fixed = QtWidgets.QHBoxLayout()
        self.la_fixed.setContentsMargins(0, 0, 0, 0)
        self.la_custom = QtWidgets.QHBoxLayout()
        self.la_custom.setContentsMargins(0, 0, 0, 0)
        self.la_preset_save = QtWidgets.QHBoxLayout()
        self.la_preset_save.setContentsMargins(0, 0, 0, 0)

        self.la_fixed.addWidget(self.rb_fixed)
        self.la_fixed.addWidget(self.sw_fixed)
        self.la_main.addLayout(self.la_fixed, 0, 0, 1, 2)
        self.la_custom.addWidget(self.rb_custom)
        self.la_main.addLayout(self.la_custom, 1, 0, 1, 2)
        self.la_main.addWidget(self.tw_cutoff, 2, 0, 1, 2)
        self.la_main.addWidget(self.cb_preset, 3, 0, 1, 1)
        self.la_main.addWidget(self.pb_history, 3, 1, 1, 1)
        self.la_preset_save.addWidget(self.cb_preset_save)
        self.la_preset_save.addWidget(self.le_preset_save)
        self.la_main.addLayout(self.la_preset_save, 4, 0, 1, 2)

        self.setLayout(self.la_main)

        self.rb_fixed.click()

    def _init_cutoff_radii(self, cutoff_radii=None):
        elements = self._radii.keys()
        if cutoff_radii is None:
            filtered_history = cutoff_history.filtered_history(
                elements,
                preferred_filenames_with_frames=self._preferred_filenames_with_frames,
            )
            if not filtered_history:
                cutoff_radii = config.Computation.std_cutoff_radius
            else:
                cutoff_radii = filtered_history[0].radii
        self.cutoff_radii = cutoff_radii

    def rb_fixed_clicked(self):
        self.sw_fixed.setCurrentIndex(0)
        if self._selected_radii_type != RadiiWidget.RadiiType.FIXED and self._discard_preset_choice_on_next_rb_click:
            self.cb_preset.discard_preset_choice()
            self._selected_radii_type = RadiiWidget.RadiiType.FIXED
        self._discard_preset_choice_on_next_rb_click = True

    def rb_custom_clicked(self):
        self.sw_fixed.setCurrentIndex(1)
        if self._selected_radii_type != RadiiWidget.RadiiType.CUSTOM and self._discard_preset_choice_on_next_rb_click:
            self.cb_preset.discard_preset_choice()
            self._selected_radii_type = RadiiWidget.RadiiType.CUSTOM
        self._discard_preset_choice_on_next_rb_click = True

    def le_fixed_text_edited(self):
        self.cb_preset.discard_preset_choice()

    def tw_cutoff_text_edited(self):
        self.cb_preset.discard_preset_choice()
        self.rb_custom.click()
        self.cb_preset.discard_preset_choice()

    def pb_history_clicked(self):
        cutoff_history_dialog = CutoffHistoryDialog(self, self._radii.keys(), self._preferred_filenames_with_frames)
        return_value = cutoff_history_dialog.exec_()
        if return_value == QtWidgets.QDialog.Rejected:
            return
        self._init_cutoff_radii(cutoff_history_dialog.selected_radii)

    def le_preset_save_changed(self, text):
        self.cb_preset_save.setChecked(text.strip() != "")

    def cb_preset_selected(self, selected_preset):
        self.cutoff_radii = selected_preset.radii

    @property
    def save_preset_name(self):
        if not self.cb_preset_save.isChecked():
            return None
        preset_name = self.le_preset_save.text().strip()
        if not re.match("[a-zA-Z0-9_]+", preset_name):
            message = QtWidgets.QMessageBox()
            message.setStandardButtons(QtWidgets.QMessageBox.Ok)
            message.setText(
                "{} is not valid preset name. Only A-Z, a-z, 0-9 and _ are valid characters".format(preset_name)
            )
            message.exec_()
            return None
        return preset_name

    @property
    def cutoff_radii(self):
        if self.rb_fixed.isChecked() and hasattr(self, "le_fixed"):
            try:
                radius_as_text = self.le_fixed.text()
                radius = float(radius_as_text)
                return radius
            except ValueError:
                message = QtWidgets.QMessageBox()
                message.setStandardButtons(QtWidgets.QMessageBox.Ok)
                message.setText("Incorrect input: {}".format(radius_as_text))
                message.exec_()
                raise

        if self.rb_custom.isChecked():
            elem_to_cutoff_radius = {}
            for i, elem in enumerate(self._radii):
                try:
                    current_radius_as_text = self.tw_cutoff.cellWidget(i, 1).text()
                    current_radius = float(current_radius_as_text)
                    elem_to_cutoff_radius[elem] = current_radius
                except ValueError:
                    message = QtWidgets.QMessageBox()
                    message.setStandardButtons(QtWidgets.QMessageBox.Ok)
                    message.setText("Incorrect input: {}".format(current_radius_as_text))
                    message.exec_()
                    raise
            return elem_to_cutoff_radius

    @cutoff_radii.setter
    def cutoff_radii(self, cutoff_radii):
        elements = self._radii.keys()
        if not isinstance(cutoff_radii, collections.abc.Iterable):
            cutoff_radii = dict((elem, float(cutoff_radii)) for elem in elements)
        # Check if all cutoff radii are equal
        if len(set(cutoff_radii.values())) == 1:
            self._discard_preset_choice_on_next_rb_click = False
            self.rb_fixed.click()
            self.le_fixed.setText(str(list(cutoff_radii.values())[0]))
        else:
            self._discard_preset_choice_on_next_rb_click = False
            self.rb_custom.click()
            for i, elem in enumerate(self._radii):
                self.tw_cutoff.cellWidget(i, 1).setText(str(cutoff_radii[elem]))



class PosBoolType(object):
    def __init__(self, pos):
        self.pos = pos

    def __getitem__(self, item):
        return self.pos[item]

    def __bool__(self):
        return bool(self.pos[0] == -1)

    def __str__(self):
        return "(%d, %d, %d)" % self.pos



# -*- coding: utf-8 -*-

__all__ = [
    "print_message",
    "progress",
    "finish",
    "error",
    "log",
    "set_output_callbacks",
]


_print_message = None
_progress = None
_finish = None
_error = None
_log = None

# Buffer log messages if no valid log callback is set
_log_buffer = []


def print_message(*args):
    if callable(_print_message):
        return _print_message(*args)


def progress(*args):
    if callable(_progress):
        return _progress(*args)


def finish(*args):
    if callable(_finish):
        return _finish(*args)


def error(*args):
    if callable(_error):
        return _error(*args)


def log(*args):
    if callable(_log):
        return _log(*args)
    else:
        _log_buffer.append(args)


def set_output_callbacks(progress_func, print_func, finished_func, error_func, log_func):
    global _progress, _print_message, _finish, _error, _log

    _progress = progress_func
    _print_message = print_func
    _finish = finished_func
    _error = error_func
    _log = log_func

    if _log_buffer and _log:
        for log_args in _log_buffer:
            _log(*log_args)
        del _log_buffer[:]



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import numpy as np
import sys
from PySide6 import QtCore, QtWidgets

try:
    from PySide6.QtOpenGLWidgets import QOpenGLWidget as QGLWidget

    has_qopenglwidget = True
except ImportError:
    from PySide6 import QtOpenGL
    from PySide6.QtOpenGLWidgets import QOpenGLWidget as QGLWidget

    has_qopenglwidget = False
from ..config.configuration import config
from OpenGL.GL import (
    GL_DEPTH_BITS,
    GL_DEPTH_BUFFER_BIT,
    GL_DEPTH_TEST,
    GL_FRAMEBUFFER,
    GL_FRAMEBUFFER_BINDING,
    GL_FRAMEBUFFER_COMPLETE,
    GL_VIEWPORT,
    glBindFramebuffer,
    glCheckFramebufferStatus,
    glClear,
    glEnable,
    glIsEnabled,
    glReadPixels,
    GL_FLOAT,
    GL_DEPTH_COMPONENT,
    glGetIntegerv,
)

from ..util.gl_util import create_perspective_projection_matrix, create_look_at_matrix


class UpdateGLEvent(QtCore.QEvent):
    def __init__(self):
        t = QtCore.QEvent.registerEventType()
        super().__init__(QtCore.QEvent.Type(t))


class GLWidget(QGLWidget if has_qopenglwidget else QGLWidget):
    """
    OpenGL widget to show the 3D-scene
    """

    def __init__(self, parent, main_window):
        if has_qopenglwidget:
            super().__init__(parent)
        else:
            super().__init__(QtOpenGL.QGLFormat(QtOpenGL.QGL.SampleBuffers), parent)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.update_needed = False
        self.dataset_loaded = False
        self.control = parent.control
        self.main_window = main_window
        self.setDisabled(True)

    @property
    def vis(self):
        return self.control.visualization

    def initializeGL(self):
        self.vis.assign_opengl_context(self)
        glEnable(GL_DEPTH_TEST)

    def minimumSizeHint(self):
        return QtCore.QSize(400, 400)

    def sizeHint(self):
        return QtCore.QSize(config.OpenGL.gl_window_size[0], config.OpenGL.gl_window_size[1])

    def mouseMoveEvent(self, e):
        dx = e.x() - self.x
        dy = e.y() - self.y
        if e.buttons() & QtCore.Qt.LeftButton:
            if e.modifiers() == QtCore.Qt.AltModifier:
                self.vis.translate_mouse(dx, dy)
            else:
                self.vis.rotate_mouse(dx, dy)
        self.x = e.x()
        self.y = e.y()
        self.updateGL()

    def wheelEvent(self, e):
        self.update_needed = True
        rot_v = 0.1
        if e.modifiers() == QtCore.Qt.ShiftModifier:
            if (e.angleDelta().x() != 0) or (e.angleDelta().y() != 0):
                self.vis.rotate_mouse(-e.angleDelta().x() * rot_v, -e.angleDelta().y() * rot_v)
        else:
            if e.angleDelta().y() != 0:
                self.vis.zoom(-e.angleDelta().y())

        QtWidgets.QApplication.postEvent(self, UpdateGLEvent())

    def mousePressEvent(self, e):
        if e.buttons() and QtCore.Qt.LeftButton:
            self.x = e.x()
            self.y = e.y()

    def mouseDoubleClickEvent(self, e):
        if e.buttons() and QtCore.Qt.LeftButton:
            x = e.x()
            y = self.height() - e.y()
            glBindFramebuffer(GL_FRAMEBUFFER, 1)  # Bind to correct framebuffer for depth check
            z = glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT)
            obj = self.vis.get_object_at_2dposition(x, y)
            if obj is None:
                x /= 1.0 * self.width()
                y /= 1.0 * self.height()
                x = x * 2 - 1
                y = y * 2 - 1
                z = z[0][0]
                z = 2 * z - 1
                z = self.vis.proj_mat[2, 3] / (-z - self.vis.proj_mat[2, 2])
                x = -x * z / self.vis.proj_mat[0, 0]
                y = -y * z / self.vis.proj_mat[1, 1]
                z += self.vis.d
                x, y, z, w = np.dot(self.vis.mat, np.array((x, y, z, 1)))
                obj = self.vis.get_object_at_3dposition(x, y, z)
            if obj is not None:
                object_type, object_index = obj
                if object_type == "atom":
                    self.window().statistics_dock.statistics_tab.html_view.show_atom(object_index)
                elif object_type == "domain":
                    self.window().statistics_dock.statistics_tab.html_view.show_domain(object_index)
                elif object_type == "surface cavity":
                    self.window().statistics_dock.statistics_tab.html_view.show_surface_cavity(object_index)
                elif object_type == "center cavity":
                    self.window().statistics_dock.statistics_tab.html_view.show_center_cavity(object_index)
                self.window().statistics_dock.raise_()
            self.updateGL()

    def customEvent(self, e):
        if self.update_needed:
            self.setDisabled(False)
            self.updateGL()
            self.update_needed = False

    #     def create_scene(self, show_box, show_atoms, show_domains, show_cavities=True, center_based_cavities=False):
    def create_scene(self):
        self.vis.create_scene()
        self.updateGL()

    def keyPressEvent(self, e):
        """
        Catches and processes key presses
        """
        rot_v_key = 15
        if e.key() == QtCore.Qt.Key_Right:
            self.vis.rotate_mouse(rot_v_key, 0)
        elif e.key() == QtCore.Qt.Key_Left:
            self.vis.rotate_mouse(-rot_v_key, 0)
        elif e.key() == QtCore.Qt.Key_Up:
            self.vis.rotate_mouse(0, -rot_v_key)
        elif e.key() == QtCore.Qt.Key_Down:
            self.vis.rotate_mouse(0, rot_v_key)
        elif e.key() == QtCore.Qt.Key_D:  # Domains
            self.vis.settings.show_domains = True
            self.vis.settings.show_surface_cavities = False
            self.vis.settings.show_center_cavities = False
            self.main_window.view_dock.view_tab.domain_check.setChecked(True)
            # self.vis.create_scene()
        elif e.key() == QtCore.Qt.Key_S:  # Cavities
            self.vis.settings.show_domains = False
            self.vis.settings.show_surface_cavities = True
            self.vis.settings.show_center_cavities = False
            self.main_window.view_dock.view_tab.surface_cavity_check.setChecked(True)
            # self.vis.create_scene()
        elif e.key() == QtCore.Qt.Key_C:  # center based cavities
            self.vis.settings.show_domains = False
            self.vis.settings.show_surface_cavities = False
            self.vis.settings.show_center_cavities = True
            self.main_window.view_dock.view_tab.center_cavity_check.setChecked(True)
            # self.vis.create_scene()
        if e.key() == QtCore.Qt.Key_R:
            self.vis.reset_view()
        else:
            e.ignore()
        self.updateGL()

    def paintGL(self):
        """
        Refresh scene
        """
        glClear(GL_DEPTH_BUFFER_BIT)
        self.vis.paint(self.width(), self.height(), has_qopenglwidget, self.devicePixelRatio())

    def updateGL(self):
        if has_qopenglwidget:
            self.update()
        else:
            super(GLWidget, self).updateGL()

    def activate(self):
        pass

    def updatestatus(self):
        self.update_needed = True



import os
import sys

from ..core import file
from ..core.control import Control

from .cli import Cli

__all__ = ["Cli"]


def main():
    file.SEARCH_PATH = os.getcwd()
    os.chdir(os.path.abspath(os.path.dirname(__file__)))
    control = Control()
    instance = Cli(control)
    instance.start()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import sys
import os
import inspect
from datetime import datetime

__all__ = ["Logger"]


class Logger(object):

    EMERG = 0
    ALERT = 1
    CRIT = 2
    ERR = 3
    ERROR = 3
    WARNING = 4
    WARN = 4
    NOTICE = 5
    INFO = 6
    DEBUG = 7

    severities = [
        "Emergency",
        "Alert",
        "Critical",
        "Error",
        "Warning",
        "Notice",
        "Informational",
        "Debug",
    ]

    def __init__(self, identifier):
        self.identifier = identifier
        self.streams = dict()
        self.format = "{time} {severity} from {identifier} in {function} ({file}:{line}): "

    def setstream(self, ident, stream, severity):
        if severity is None or stream == -1:
            if ident in self.streams:
                del self.streams[ident]
        else:
            self.streams[ident] = (stream, severity)

    def log(self, severtiy, message):
        self._log(severtiy, message)

    def emerg(self, message):
        self._log(self.EMERG, message)

    def alert(self, message):
        self._log(self.ALERT, message)

    def crit(self, message):
        self._log(self.CRIT, message)

    def err(self, message):
        self._log(self.ERR, message)

    def error(self, message):
        self._log(self.ERR, message)

    def warn(self, message):
        self._log(self.WARN, message)

    def notice(self, message):
        self._log(self.NOTICE, message)

    def info(self, message):
        self._log(self.INFO, message)

    def debug(self, message):
        self._log(self.DEBUG, message)

    def _log(self, severity, message):
        st = inspect.stack()
        info = {
            "identifier": self.identifier,
            "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "severity": self.severities[severity].upper(),
            "function": st[2][3],
            "file": os.path.relpath(st[2][1]),
            "line": st[2][2],
        }
        line = self.format.format(**info) + message
        self._logline(severity, line)

    def _logline(self, severity, line):
        for stream, maxseverity in self.streams.values():
            if severity <= maxseverity:
                if isinstance(stream, Logger):
                    stream._logline(severity, line)
                else:
                    #  print >>stream, line
                    print(line, file=stream)



from PySide6 import QtCore, QtGui
import gl_widget
import main_window

ignore_list = [
    QtCore.QEvent.Type.HideToParent,
    QtCore.QEvent.Type.DynamicPropertyChange,
    QtCore.QEvent.Type.ChildRemoved,
    QtCore.QEvent.Type.Destroy,
    QtCore.QEvent.Type.WinIdChange,
    QtCore.QEvent.Type.ActionChanged,
    QtCore.QEvent.Type.Hide,
    QtCore.QEvent.Type.ShortcutOverride,
    QtCore.QEvent.Type.KeyPress,
    QtCore.QEvent.Type.Timer,
    QtCore.QEvent.Type.Close,
    QtCore.QEvent.Type.WindowDeactivate,
    QtCore.QEvent.Type.FocusIn,
    QtCore.QEvent.Type.Style,
    QtCore.QEvent.Type.ApplicationActivate,
    QtCore.QEvent.Type.Paint,
    QtCore.QEvent.Type.FocusOut,
    QtCore.QEvent.Type.DeferredDelete,
    QtCore.QEvent.Type.Resize,
    QtCore.QEvent.Type.Move,
    QtCore.QEvent.Type.LayoutRequest,
    QtCore.QEvent.Type.ChildAdded,
    QtCore.QEvent.Type.Show,
    QtCore.QEvent.Type.ShowToParent,
    QtCore.QEvent.Type.WindowActivate,
    QtCore.QEvent.Type.Polish,
    QtCore.QEvent.Type.ChildPolished,
    QtCore.QEvent.Type.PolishRequest,
    QtCore.QEvent.Type.MetaCall,
    QtCore.QEvent.Type.MacGLWindowChange,
    QtCore.QEvent.Type.ActivationChange,
    QtCore.QEvent.Type.WindowStateChange,
    QtCore.QEvent.Type.ParentChange,
    QtCore.QEvent.Type.Create,
    QtCore.QEvent.Type.MacGLClearDrawable,
    QtCore.QEvent.Type.FontChange,
    QtCore.QEvent.Type.PaletteChange,
    QtCore.QEvent.Type.HoverEnter,
    QtCore.QEvent.Type.Enter,
    QtCore.QEvent.Type.HoverMove,
    QtCore.QEvent.Type.MouseButtonRelease,
    QtCore.QEvent.Type.MouseTrackingChange,
    QtCore.QEvent.Type.CursorChange,
    QtCore.QEvent.Type.WindowTitleChange,
    QtCore.QEvent.Type.ContentsRectChange,
    QtCore.QEvent.Type.AcceptDropsChange,
    QtCore.QEvent.Type.MouseTrackingChange,
    QtCore.QEvent.Type.ZOrderChange,
    QtCore.QEvent.Type.MouseMove,
    QtCore.QEvent.Type.Leave,
    QtCore.QEvent.Type.HoverLeave,
    QtCore.QEvent.Type.ApplicationDeactivated,
    QtCore.QEvent.Type.EnabledChange,
    QtCore.QEvent.Type.WindowBlocked,
    QtCore.QEvent.Type.EnabledChange,
    QtCore.QEvent.Type.WindowUnblocked,
    QtCore.QEvent.Type.KeyRelease,
    QtCore.QEvent.Type.NativeGesture,
    QtCore.QEvent.Type.KeyboardLayoutChange,
    QtCore.QEvent.Type.ToolTip,
]
test = QtCore.QEvent.registerEventType()
t = QtCore.QEvent.Type(test)


class UpdateGLEvent(QtCore.QEvent):

    def __init__(self):
        global t
        QtCore.QEvent.__init__(self, t)


class MyApplication(QtGui.QApplication):
    def __init__(self, args):
        QtGui.QApplication.__init__(self, args)
        self.installEventFilter(self)

    def eventFilter(self, receiver, event):
        global ignore_list
        # print(receiver, event.type(), t)
        global t
        if event.type() == QtCore.QEvent.MouseMove and isinstance(receiver, gl_widget.GLWidget):
            if receiver.dataset_loaded:
                if event.buttons() & QtCore.Qt.LeftButton:
                    dx = event.x() - receiver.x
                    dy = event.y() - receiver.y
                    receiver.vis.rotate_mouse(dx, dy)
                    receiver.x = event.x()
                    receiver.y = event.y()
                    receiver.updateGL()
            return True

        if event.type() == QtCore.QEvent.Wheel and isinstance(receiver, gl_widget.GLWidget):
            if receiver.dataset_loaded:
                receiver.update_needed = True
                if event.modifiers() != QtCore.Qt.ShiftModifier:
                    if event.orientation() == QtCore.Qt.Orientation.Vertical:
                        receiver.vis.zoom(event.delta())
                else:
                    rot_v = 0.1
                    if event.orientation() == QtCore.Qt.Orientation.Horizontal:
                        receiver.vis.rotate_mouse(event.delta() * rot_v, 0)
                    else:
                        receiver.vis.rotate_mouse(0, event.delta() * rot_v)

                QtGui.QApplication.postEvent(receiver, UpdateGLEvent())
                return True

        if event.type() == QtCore.QEvent.MouseButtonPress and isinstance(receiver, gl_widget.GLWidget):
            if receiver.dataset_loaded:
                if event.buttons() and QtCore.Qt.LeftButton:
                    receiver.x = event.x()
                    receiver.y = event.y()
                    return True

        if event.type() == t and isinstance(receiver, gl_widget.GLWidget):
            if receiver.update_needed:
                receiver.updateGL()
                receiver.update_needed = False
            return True

        if event.type() == QtCore.QEvent.KeyPress and isinstance(receiver, gl_widget.GLWidget):
            if receiver.dataset_loaded:
                rot_v_key = 15
                if event.key() == QtCore.Qt.Key_Right:
                    receiver.vis.rotate_mouse(rot_v_key, 0)
                elif event.key() == QtCore.Qt.Key_Left:
                    receiver.vis.rotate_mouse(-rot_v_key, 0)
                elif event.key() == QtCore.Qt.Key_Up:
                    receiver.vis.rotate_mouse(0, -rot_v_key)
                elif event.key() == QtCore.Qt.Key_Down:
                    receiver.vis.rotate_mouse(0, rot_v_key)
                elif event.key() == QtCore.Qt.Key_D:  # Domains
                    receiver.vis.create_scene(False)
                elif event.key() == QtCore.Qt.Key_C:  # Cavities
                    receiver.vis.create_scene(True)
                elif event.key() == QtCore.Qt.Key_F:  # center based cavities
                    receiver.vis.create_scene(True, True)
                else:
                    event.ignore()
                receiver.updateGL()
                return True

        if event.type() == QtCore.QEvent.KeyPress and isinstance(receiver, main_window.MainWindow):
            if event.key() == QtCore.Qt.Key_M:
                if not self.isFullScreen():
                    for dock in self.docks:
                        dock.hide()
                    self.showFullScreen()
                else:
                    for dock in self.docks:
                        dock.show()
                    self.showNormal()
                return True

        if event.type() not in ignore_list:
            print(event.type(), event)
        return super(MyApplication, self).eventFilter(receiver, event)



class ProhibitedError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


def prohibited(f):
    raise ProhibitedError("Function call is prohibited.")



from PySide6 import QtCore, QtWidgets


class CalculationTable(QtWidgets.QTableView):
    def __init__(self, parent):
        QtWidgets.QTableView.__init__(self, parent)
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)


class TableModel(QtCore.QAbstractTableModel):
    def __init__(self, parent, data, header):
        QtCore.QAbstractTableModel.__init__(self, parent)
        self.databla = list(data)
        self.header = header

    def rowCount(self, parent):
        return len(self.databla)

    def columnCount(self, parent):
        return len(self.databla[0])

    #
    def data(self, index, role):
        if not index.isValid():
            return None
        elif role != QtCore.Qt.DisplayRole:
            return None
        return self.databla[index.row()][index.column()]

    def headerData(self, col, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self.header[col]
        return None


#    def sort(self, col, order):
#        """sort table by given column number col"""
#        self.emit(SIGNAL("layoutAboutToBeChanged()"))
#        self.mylist = sorted(self.mylist,
#            key=operator.itemgetter(col))
#        if order == Qt.DescendingOrder:
#            self.mylist.reverse()
#        self.emit(SIGNAL("layoutChanged()"))



# -*- coding: utf-8 -*-

from PySide6 import QtWidgets
from PySide6.QtCore import QTimer, QProcess
import gr3
from ..gl_widget import GLWidget
from ...core import file
import os.path
import shutil
import sys
import tempfile


class ImageVideoTabDock(QtWidgets.QDockWidget):
    """
    DockWidget for the 'image/video'-tab
    """

    def __init__(self, parent):
        super().__init__("image/video", parent)  # self?
        self.setWidget(QtWidgets.QWidget())

        self.layout = QtWidgets.QHBoxLayout()
        self.image_video_tab = ImageVideoTab(self.widget(), parent)

        self.layout.addWidget(self.image_video_tab)
        self.widget().setLayout(self.layout)

        self.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable | QtWidgets.QDockWidget.DockWidgetFloatable)
        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)


class ImageVideoTab(QtWidgets.QWidget):
    """
    tab 'image/video' in the main widget
    """

    def __init__(self, parent, main_window):
        super().__init__(parent)
        self.main_window = main_window
        self.init_gui()

    def init_gui(self):
        self.vbox = QtWidgets.QVBoxLayout()

        self.screenshot_button = QtWidgets.QPushButton("Save screenshot for the current frame", self)
        self.video_button = QtWidgets.QPushButton("Save video for all selected frames", self)
        self.mass_screenshot_button = QtWidgets.QPushButton("Save screenshot for all selected frames", self)

        self.screenshot_button.clicked.connect(self.save_screenshot)
        self.screenshot_button.setDisabled(True)
        self.mass_screenshot_button.clicked.connect(self.save_screenshot_for_all_selected_frames)
        self.mass_screenshot_button.setDisabled(True)
        self.video_button.clicked.connect(self.save_video_for_all_selected_frames)
        self.video_button.setDisabled(True)

        self.vbox.addWidget(self.screenshot_button)
        self.vbox.addWidget(self.mass_screenshot_button)
        self.vbox.addWidget(self.video_button)
        self.vbox.addStretch()
        self.setLayout(self.vbox)

        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)

    def save_screenshot(self):
        file_name = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save screenshot...", filter="Portable Network Graphics (*.png)"
        )[0]
        if file_name:
            ext = os.path.splitext(file_name)[1]
            if not ext:
                file_name += ".png"
            self.main_window.control.visualization.save_screenshot(file_name)

    def get_selected_frames(self):
        file_list = self.main_window.file_dock.file_tab.file_list
        selection = file_list.get_selection()
        if not selection:
            QtWidgets.QMessageBox.information(
                self,
                "No frame selected",
                "Please use the file tab to select at least one frame.",
                QtWidgets.QMessageBox.Ok,
                QtWidgets.QMessageBox.NoButton,
            )
            return

        frames_to_write = []
        for file_name, frame_numbers in selection.items():
            if frame_numbers == [-1]:
                num_frames = file.File.open(file_name).info.num_frames
                frame_numbers = range(num_frames)
            for frame_number in frame_numbers:
                frames_to_write.append((file_name, frame_number))
        return frames_to_write

    def save_video_for_all_selected_frames(self):
        frames_to_write = self.get_selected_frames()
        if not frames_to_write:
            return
        file_name = QtWidgets.QFileDialog.getSaveFileName(self, "Save video...", filter="QuickTime Movie (*.mov)")[0]
        if file_name:
            ext = os.path.splitext(file_name)[1]
            if not ext:
                file_name += ".mov"
            dialog = MassScreenshotAndVideoDialog(
                self, frames_to_write, should_save_video=True, video_file_name=file_name
            )
            dialog.show()
            return

    def save_screenshot_for_all_selected_frames(self):
        frames_to_write = self.get_selected_frames()
        if not frames_to_write:
            return
        dir_name = QtWidgets.QFileDialog.getExistingDirectory(self, "Save screenshots to folder...")
        if dir_name:
            dialog = MassScreenshotAndVideoDialog(self, frames_to_write, dir_name)
            dialog.show()
            return


class MassScreenshotAndVideoDialog(QtWidgets.QDialog):
    def __init__(
        self,
        parent,
        frames_to_write,
        dir_name=None,
        should_save_video=False,
        video_file_name=None,
    ):
        super().__init__(parent)
        self.control = parent.main_window.control
        self.setModal(True)
        self.frames_to_write = frames_to_write
        self.images_written = []
        if not dir_name:
            dir_name = tempfile.mkdtemp()
            self.delete_dir = True
        else:
            self.delete_dir = False
        self.dir_name = dir_name
        self.should_save_video = should_save_video
        self.video_file_name = video_file_name
        self.vbox = QtWidgets.QVBoxLayout()
        self.progress_bar = QtWidgets.QProgressBar(self)
        self.progress_bar.setValue(0)
        self.progress_bar.setMinimum(0)
        if self.should_save_video:
            self.progress_bar.setMaximum(int(2.5 * len(frames_to_write)))
        else:
            self.progress_bar.setMaximum(len(frames_to_write))
        self.vbox.addWidget(self.progress_bar)
        self.setLayout(self.vbox)
        self.is_rejected = False
        self.process = None
        self.is_first_frame = True
        if self.should_save_video:
            self.setWindowTitle("Saving video...")
        else:
            self.setWindowTitle("Saving screenshots...")
        QTimer.singleShot(10, self.save_single_screenshot)

    def save_single_screenshot(self):
        if self.is_rejected:
            return
        current_frame = self.frames_to_write.pop(0)
        file_name, frame_number = current_frame
        width, height = 1920, 1080
        image_file_name = os.path.join(
            self.dir_name,
            "{}.{:06d}.png".format(os.path.basename(file_name), frame_number + 1),
        )
        self.control.visualize(file_name, frame_number)
        self.control.visualization.create_scene()
        self.control.visualization.save_screenshot(
            image_file_name,
            width,
            height,
            self.is_first_frame,
            not self.frames_to_write,
        )
        self.images_written.append(image_file_name)
        self.is_first_frame = False
        if self.frames_to_write:
            self.progress_bar.setValue(len(self.images_written))
            self.update()
            QTimer.singleShot(10, self.save_single_screenshot)
        elif self.should_save_video:
            QTimer.singleShot(10, self.save_video)
        else:
            self.done(0)

    def save_video(self):
        self.process = QProcess()
        self.process.start(
            "gr",
            ["../util/video_output.py", self.video_file_name] + self.images_written,
        )
        self.process.readyReadStandardOutput.connect(self.process_output)
        self.process.readyReadStandardError.connect(self.process_error)
        self.process.finished.connect(lambda *args: self.finished_video())

    def finished_video(self):
        if self.delete_dir:
            shutil.rmtree(self.dir_name)
        self.done(0)

    def process_output(self):
        output = self.process.readAllStandardOutput()
        for line in str(output).splitlines():
            try:
                number = int(line)
            except ValueError:
                pass
            else:
                self.progress_bar.setValue(len(self.images_written) + number)

    def process_error(self):
        output = self.process.readAllStandardError()
        sys.stderr.write(str(output, "utf-8"))

    def reject(self):
        self.is_rejected = True
        if self.process is not None:
            self.process.terminate()
        super(MassScreenshotAndVideoDialog, self).reject()



import sys


def excepthook(type, value, tb):
    import traceback
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import TerminalFormatter
    from pygments.token import Token

    tbtext = "".join(traceback.format_exception(type, value, tb))

    lexer = get_lexer_by_name("pytb", stripall=True)

    formatter = TerminalFormatter(bg="dark")  # dark or light
    formatter.colorscheme[Token.Literal.Number] = ("yellow", "yellow")

    sys.stderr.write(highlight(tbtext, lexer, formatter))


sys.excepthook = excepthook



from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QTimer
from .statistics.tree_list import TreeList
from .statistics.html_view import HTMLWindow


class StatisticsTabDock(QtWidgets.QDockWidget):
    """
    DockWidget for the 'statistics'-tab
    """

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, "statistics", parent)
        self.setWidget(QtWidgets.QWidget())

        self.layout = QtWidgets.QHBoxLayout()
        self.statistics_tab = StatisticsTab(self.widget())

        self.layout.addWidget(self.statistics_tab)
        self.widget().setLayout(self.layout)

        self.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable | QtWidgets.QDockWidget.DockWidgetFloatable)
        self.setVisible(False)
        self.setMinimumWidth(350)

    def update_results(self, results):
        if results is not None:
            self.setVisible(True)
            self.statistics_tab.update_results(results)


class StatisticsTab(QtWidgets.QWidget):

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)
        self.control = None
        hbox = QtWidgets.QHBoxLayout()

        self.html_view = HTMLWindow()
        self.tree_list = TreeList(self.html_view)

        self.html_view.tree_list = self.tree_list
        self.tree_list.setMinimumWidth(150)
        hbox.addWidget(self.tree_list)
        hbox.addWidget(self.html_view)
        self.setLayout(hbox)
        hbox.setContentsMargins(0, 0, 0, 0)
        self.show()

    def update_results(self, results):
        self.html_view.update_results(results)
        self.tree_list.update_results(results)



# -*- coding: utf-8 -*-

from __future__ import absolute_import


from PySide6 import QtCore, QtGui, QtWidgets


class LabeledComboBox(QtWidgets.QComboBox):

    class Delegate(QtWidgets.QItemDelegate):
        def __init__(self, parent):
            super().__init__(parent)

        def paint(self, painter, option, index):
            role = index.data(QtCore.Qt.AccessibleDescriptionRole)
            if role == "parent":
                option.state |= QtWidgets.QStyle.State_Enabled
            else:
                indent = option.fontMetrics.width(2 * " ")
                option.rect.adjust(indent, 0, 0, 0)
                option.textElideMode = QtCore.Qt.ElideNone
            super(LabeledComboBox.Delegate, self).paint(painter, option, index)

    def __init__(self, label, parent=None):
        super().__init__(parent)
        self._label = label
        LabeledComboBox._init_ui(self)  # Emulate static binding

    def _init_ui(self):
        self.setItemDelegate(LabeledComboBox.Delegate(self))
        label_item = QtGui.QStandardItem(self._label)
        label_item.setFlags(label_item.flags() & ~(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable))
        label_item.setData("parent", QtCore.Qt.AccessibleDescriptionRole)
        label_item_font = label_item.font()
        label_item_font.setBold(True)
        label_item.setFont(label_item_font)
        self.model().appendRow(label_item)



# -*- coding: utf-8 -*-


__all__ = ["atomstogrid", "mark_cavities", "cavity_triangles", "cavity_intersections"]


# Import gr3 to load `libGR3.so` which is needed by the ctypes extension
import gr3  # noqa: F401 pylint: disable=unused-import
import numpy as np
from ctypes import (
    c_int,
    c_int64,
    c_int8,
    POINTER,
    sizeof,
    CDLL,
    Structure,
    byref,
    c_float,
    cast,
)
import sys
import os
import platform


int_type = np.dtype(c_int)


class subgrid_cell_t(Structure):
    _fields_ = [
        ("num_atoms", c_int),
        ("atom_positions", POINTER(c_int)),
        ("num_domains", c_int),
        ("domain_indices", POINTER(c_int)),
        ("domain_positions", POINTER(c_int)),
    ]


class subgrid_t(Structure):
    _fields_ = [
        ("subgrid_cell_t", POINTER(subgrid_cell_t)),
        ("cubesize", c_int),
        ("ncells", c_int),
        ("dimensions", c_int * 3),
        ("strides", c_int * 3),
    ]


shared_lib_prefix = {
    "Windows": "",
}.get(platform.system(), "lib")

shared_lib_ext = {
    "Darwin": ".dylib",
    "Windows": ".dll",
}.get(platform.system(), ".so")

libpath = os.path.join(
    os.path.realpath(os.path.dirname(__file__)),
    shared_lib_prefix + "algorithm" + shared_lib_ext,
)
lib = CDLL(libpath)

lib.atomstogrid.restype = None
lib.atomstogrid.argtypes = [
    POINTER(c_int64),  # grid
    c_int * 3,  # dimensions
    c_int * 3,  # strides
    c_int,  # natoms
    POINTER(c_int),  # atom_positions
    POINTER(c_int),  # radii_indices
    c_int,  # nradii
    POINTER(c_int),  # radii
    c_int,  # ntranslations
    POINTER(c_int),  # translations
    POINTER(c_int8),  # discretization_grid
    c_int * 3,
]  # discretization_grid_strides

lib.subgrid_create.restype = POINTER(subgrid_t)
lib.subgrid_create.argtypes = [c_int, c_int * 3]  # cubesize  # grid_dimensions

lib.subgrid_destroy.restype = None
lib.subgrid_destroy.argtypes = [POINTER(subgrid_t)]  # sg

lib.subgrid_add_atoms.restype = None
lib.subgrid_add_atoms.argtypes = [
    POINTER(subgrid_t),  # sg
    c_int,  # natoms
    POINTER(c_int),  # atom_positions
    c_int,  # ntranslations
    POINTER(c_int),
]  # translations

lib.subgrid_add_domains.restype = None
lib.subgrid_add_domains.argtypes = [
    POINTER(subgrid_t),  # sg
    c_int,  # npoints
    POINTER(c_int),  # domain_indices
    POINTER(c_int),  # domain_points
    c_int,  # ntranslations
    POINTER(c_int),
]  # translations

lib.mark_cavities.restype = None
lib.mark_cavities.argtypes = [
    POINTER(c_int64),  # grid
    POINTER(c_int64),  # domain_grid
    c_int * 3,  # dimensions
    c_int * 3,  # strides
    POINTER(c_int8),  # discretization_grid
    c_int * 3,  # discgrid_strides
    POINTER(subgrid_t),  # sg
    c_int,
]  # use_surface_points

lib.cavity_triangles.restype = c_int
lib.cavity_triangles.argtypes = [
    POINTER(c_int64),  # cavity_grid
    c_int * 3,  # dimensions
    c_int * 3,  # strides
    c_int,  # ncavity_indices
    POINTER(c_int),  # cavity_indices
    c_int,  # isolevel
    c_float * 3,  # step
    c_float * 3,  # offset
    POINTER(c_int8),  # discretization_grid
    c_int * 3,  # discgrid_strides
    POINTER(POINTER(c_float)),  # vertices
    POINTER(POINTER(c_float)),  # normals
    POINTER(c_float),
]  # surface_area

lib.free_float_p.restype = None
lib.free_float_p.argtypes = [POINTER(c_float)]

lib.cavity_intersections.restype = None
lib.cavity_intersections.argtypes = [
    POINTER(c_int64),  # grid
    c_int * 3,  # dimensions
    c_int * 3,  # strides
    c_int,  # num_domains
    POINTER(c_int8),
]  # intersection_table

lib.mark_translation_vectors.restype = None
lib.mark_translation_vectors.argtypes = [
    POINTER(c_int8),  # grid
    c_int * 3,  # dimensions
    c_int * 3,  # strides
    c_int,  # ntranslations
    POINTER(c_int),
]  # translations


def atomstogrid(
    grid,
    discrete_positions,
    radii_indices,
    discrete_radii,
    translation_vectors,
    discretization_grid,
):
    dimensions = (c_int * 3)(*grid.shape)
    strides = (c_int * 3)(*[s // grid.itemsize for s in grid.strides])
    grid_p = grid.ctypes.data_as(POINTER(c_int64))

    discrete_positions = np.ascontiguousarray(discrete_positions, dtype=int_type)
    natoms = c_int(discrete_positions.shape[0])
    discrete_positions_p = discrete_positions.ctypes.data_as(POINTER(c_int))

    radii_indices = np.ascontiguousarray(radii_indices, dtype=int_type)
    radii_indices_p = radii_indices.ctypes.data_as(POINTER(c_int))

    discrete_radii = np.ascontiguousarray(discrete_radii, dtype=int_type)
    nradii = discrete_radii.shape[0]
    discrete_radii_p = discrete_radii.ctypes.data_as(POINTER(c_int))

    translation_vectors = np.ascontiguousarray(translation_vectors, dtype=int_type)
    ntranslations = c_int(translation_vectors.shape[0])
    translation_vectors_p = translation_vectors.ctypes.data_as(POINTER(c_int))

    discretization_grid_strides = (c_int * 3)(*[s // discretization_grid.itemsize for s in discretization_grid.strides])
    discretization_grid_p = discretization_grid.ctypes.data_as(POINTER(c_int8))

    lib.atomstogrid(
        grid_p,
        dimensions,
        strides,
        natoms,
        discrete_positions_p,
        radii_indices_p,
        nradii,
        discrete_radii_p,
        ntranslations,
        translation_vectors_p,
        discretization_grid_p,
        discretization_grid_strides,
    )


def subgrid_create(cubesize, grid_dimensions):
    cubesize_c = c_int(cubesize)
    grid_dimensions_c = (c_int * 3)(*grid_dimensions)
    sg = lib.subgrid_create(cubesize_c, grid_dimensions_c)
    return sg


def subgrid_destroy(sg):
    lib.subgrid_destroy(sg)


def subgrid_add_atoms(sg, atom_positions, translation_vectors):
    atom_positions = np.ascontiguousarray(atom_positions, dtype=int_type)
    natoms_c = c_int(atom_positions.shape[0])
    atom_positions_c = atom_positions.ctypes.data_as(POINTER(c_int))

    translations = np.ascontiguousarray(translation_vectors, dtype=int_type)
    ntranslations_c = c_int(translations.shape[0])
    translations_c = translations.ctypes.data_as(POINTER(c_int))

    lib.subgrid_add_atoms(sg, natoms_c, atom_positions_c, ntranslations_c, translations_c)


def subgrid_add_domains(sg, domain_indices, domain_points, translation_vectors):
    domain_indices = np.ascontiguousarray(domain_indices, dtype=int_type)
    npoints_c = c_int(domain_indices.shape[0])
    domain_indices_c = domain_indices.ctypes.data_as(POINTER(c_int))

    domain_points = np.ascontiguousarray(domain_points, dtype=int_type)
    domain_points_c = domain_points.ctypes.data_as(POINTER(c_int))

    translations = np.ascontiguousarray(translation_vectors, dtype=int_type)
    ntranslations_c = c_int(translations.shape[0])
    translations_c = translations.ctypes.data_as(POINTER(c_int))

    lib.subgrid_add_domains(
        sg,
        npoints_c,
        domain_indices_c,
        domain_points_c,
        ntranslations_c,
        translations_c,
    )


def mark_cavities_c(grid, domain_grid, discretization_grid, sg, use_surface_points):
    dimensions_c = (c_int * 3)(*grid.shape)
    strides_c = (c_int * 3)(*[s // grid.itemsize for s in grid.strides])
    grid_c = grid.ctypes.data_as(POINTER(c_int64))

    if domain_grid is not None:
        domain_grid_c = domain_grid.ctypes.data_as(POINTER(c_int64))
    else:
        domain_grid_c = POINTER(c_int64)()

    discgrid_strides_c = (c_int * 3)(*[s // discretization_grid.itemsize for s in discretization_grid.strides])
    discretization_grid_c = discretization_grid.ctypes.data_as(POINTER(c_int8))

    use_surface_points_c = c_int(use_surface_points)

    lib.mark_cavities(
        grid_c,
        domain_grid_c,
        dimensions_c,
        strides_c,
        discretization_grid_c,
        discgrid_strides_c,
        sg,
        use_surface_points_c,
    )


def mark_cavities(
    domain_grid,
    discretization_grid,
    grid_dimensions,
    sg_cube_size,
    atom_positions,
    translation_vectors,
    domain_point_list,
    use_surface_points,
):

    # step 1
    sg = subgrid_create(sg_cube_size, grid_dimensions)
    # step 2
    subgrid_add_atoms(sg, atom_positions, translation_vectors)

    # step 3
    domain_indices = []
    domain_points = []
    for i in range(len(domain_point_list)):
        for p in domain_point_list[i]:
            domain_indices.append(i)
            domain_points.append(p)
    subgrid_add_domains(sg, domain_indices, domain_points, translation_vectors)

    grid = np.zeros(grid_dimensions, dtype=np.int64)
    # step 4 and 5
    mark_cavities_c(grid, domain_grid, discretization_grid, sg, use_surface_points)

    subgrid_destroy(sg)

    return grid


def cavity_triangles(cavity_grid, cavity_indices, isolevel, step, offset, discretization_grid):
    cavity_grid_c = cavity_grid.ctypes.data_as(POINTER(c_int64))
    dimensions_c = (c_int * 3)(*cavity_grid.shape)
    strides_c = (c_int * 3)(*[s // cavity_grid.itemsize for s in cavity_grid.strides])

    if not isinstance(cavity_indices, np.ndarray) and not isinstance(cavity_indices, list):
        cavity_indices = list(cavity_indices)
    ncavity_indices_c = c_int(len(cavity_indices))
    cavity_indices = np.ascontiguousarray(cavity_indices, dtype=int_type)
    cavity_indices_c = cavity_indices.ctypes.data_as(POINTER(c_int))

    isolevel_c = c_int(isolevel)
    step_c = (c_float * 3)(*step)
    offset_c = (c_float * 3)(*offset)

    discretization_grid_c = discretization_grid.ctypes.data_as(POINTER(c_int8))
    discgrid_strides_c = (c_int * 3)(*[s // discretization_grid.itemsize for s in discretization_grid.strides])

    vertices_c = POINTER(c_float)()
    normals_c = POINTER(c_float)()
    surface_area_c = c_float()

    ntriangles = lib.cavity_triangles(
        cavity_grid_c,
        dimensions_c,
        strides_c,
        ncavity_indices_c,
        cavity_indices_c,
        isolevel_c,
        step_c,
        offset_c,
        discretization_grid_c,
        discgrid_strides_c,
        byref(vertices_c),
        byref(normals_c),
        byref(surface_area_c),
    )

    ArrayType = c_float * ntriangles * 3 * 3
    vertices_p = cast(vertices_c, POINTER(ArrayType))
    vertices = np.frombuffer(vertices_p.contents, dtype=c_float)
    vertices = np.array(vertices, dtype=float, copy=True).reshape((ntriangles, 3, 3))
    lib.free_float_p(vertices_c)
    normals_p = cast(normals_c, POINTER(ArrayType))
    normals = np.frombuffer(normals_p.contents, dtype=c_float)
    normals = np.array(normals, dtype=float, copy=True).reshape((ntriangles, 3, 3))
    lib.free_float_p(normals_c)
    surface_area = surface_area_c.value

    return vertices, normals, surface_area


def cavity_intersections(grid, num_domains):
    dimensions_c = (c_int * 3)(*grid.shape)
    strides_c = (c_int * 3)(*[s // grid.itemsize for s in grid.strides])
    grid_c = grid.ctypes.data_as(POINTER(c_int64))

    num_domains_c = c_int(num_domains)
    intersection_table = np.zeros((num_domains, num_domains), dtype=np.int8)
    intersection_table_c = intersection_table.ctypes.data_as(POINTER(c_int8))

    lib.cavity_intersections(grid_c, dimensions_c, strides_c, num_domains_c, intersection_table_c)

    return intersection_table


def mark_translation_vectors(grid, translation_vectors):
    dimensions_c = (c_int * 3)(*grid.shape)
    strides_c = (c_int * 3)(*[s // grid.itemsize for s in grid.strides])
    grid_c = grid.ctypes.data_as(POINTER(c_int8))

    translation_vectors = np.ascontiguousarray(translation_vectors, dtype=int_type)
    ntranslations_c = c_int(translation_vectors.shape[0])
    translation_vectors_c = translation_vectors.ctypes.data_as(POINTER(c_int))

    lib.mark_translation_vectors(grid_c, dimensions_c, strides_c, ntranslations_c, translation_vectors_c)



# -*- coding: utf-8 -*-

from PySide6 import QtCore, QtWidgets
import os.path
import functools
from ...core import calculation, volumes, file
from ..dialogs.calc_settings_dialog import CalculationSettingsDialog
from ..dialogs.progress_dialog import ProgressDialog
from ...config.configuration import config
from ...util.message import print_message, progress, finish
from ..gl_widget import GLWidget, UpdateGLEvent
from ...core.file import File
from ...util import message


class FileTabDock(QtWidgets.QDockWidget):
    """
    DockWidget for the 'file'-tab
    """

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, "file", parent)
        self.setWidget(QtWidgets.QWidget(self))

        self.layout = QtWidgets.QHBoxLayout()
        self.file_tab = FileTab(self.widget(), parent)

        self.layout.addWidget(self.file_tab)
        self.widget().setLayout(self.layout)

        self.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable | QtWidgets.QDockWidget.DockWidgetFloatable)


class CalculationThread(QtCore.QThread):
    """
    Thread to calculate the cavities
    """

    def __init__(self, parent, func, settings):
        QtCore.QThread.__init__(self, parent)
        self.func = func
        self.settings = settings
        self._exited_with_errors = False

    @property
    def exited_with_errors(self):
        return self._exited_with_errors

    def run(self):
        try:
            self.func(self.settings)
        except Exception as e:
            self._exited_with_errors = True
            message.error(str(e))
            message.finish()
            raise


class FileTab(QtWidgets.QWidget):
    """
    tab 'file' in the main widget
    """

    def __init__(self, parent, main_window):
        QtWidgets.QWidget.__init__(self, parent)
        self.main_window = main_window
        self.progress_dialog = ProgressDialog(self)
        self.most_recent_path = "~"

        self.control = main_window.control

        self.init_gui()
        self.guessed_volumes_for = set()
        self.last_shown_filename_with_frame = None

    def init_gui(self):
        self.vbox = QtWidgets.QVBoxLayout()
        self.vbox.setSpacing(0)
        self.button_hbox = QtWidgets.QHBoxLayout()
        self.button_hbox.setSpacing(10)

        self.file_button = QtWidgets.QPushButton("Open", self)
        self.file_button.setDefault(True)
        self.file_button.clicked.connect(self.open_file_dialog)
        self.button_hbox.addWidget(self.file_button)

        self.delete_button = QtWidgets.QPushButton("Delete", self)
        self.delete_button.clicked.connect(self.remove_selected_files)
        self.delete_button.setDisabled(True)
        self.button_hbox.addWidget(self.delete_button)

        self.calculate_button = QtWidgets.QPushButton("Calculate", self)
        self.calculate_button.clicked.connect(self.calculate)
        self.calculate_button.setDisabled(True)
        self.button_hbox.addWidget(self.calculate_button)

        self.show_button = QtWidgets.QPushButton("Show", self)
        self.show_button.clicked.connect(self.show_selected_frame)
        self.show_button.setDisabled(True)
        self.button_hbox.addWidget(self.show_button)

        self.vbox.addLayout(self.button_hbox)

        self.button2_hbox = QtWidgets.QHBoxLayout()
        self.button2_hbox.setSpacing(10)

        self.select_all_button = QtWidgets.QPushButton("Select all frames", self)
        self.select_all_button.clicked.connect(self.select_all)
        self.select_all_button.setDisabled(True)
        self.button2_hbox.addWidget(self.select_all_button)

        self.select_nth_button = QtWidgets.QPushButton("Select every nth frame...", self)
        self.select_nth_button.clicked.connect(self.select_nth)
        self.select_nth_button.setDisabled(True)
        self.button2_hbox.addWidget(self.select_nth_button)

        self.vbox.addLayout(self.button2_hbox)

        #        self.file_list = DragList(self)
        #        self.file_list.itemDoubleClicked.connect(self.calculate)
        #        self.file_list.itemSelectionChanged.connect(self.selection_changed)
        #        self.vbox.addWidget(self.file_list)
        #
        #        for path in config.recent_files:
        #            self.file_list.add_file(path)

        self.file_list = TreeList(self)
        self.vbox.addSpacing(10)
        self.vbox.addWidget(self.file_list)

        self.setLayout(self.vbox)

    def select_all(self):
        self.file_list.select_all()

    def select_nth(self):
        n, okay = QtWidgets.QInputDialog.getInt(self, "Set n", "", 1, 1)
        if okay:
            self.file_list.select_nth(n)

    def show_selected_frame(self):
        self.last_shown_filename_with_frame = self.file_list.show_selected_frame()

    def selection_changed(self):
        sel = self.file_list.get_selection()
        if not sel or len(self.file_list.get_selection()) > 1:
            return

    def remove_selected_files(self):
        self.enable_files_in_menu()
        self.file_list.remove_selected_files()

    def enable_files_in_menu(self):
        actions = self.main_window.recent_files_submenu.actions()
        selected = self.file_list.selectedItems()
        if not selected:
            return
        text = None
        subitem = selected[0]
        item = subitem.parent()

        # if sheet contains only one frame and this is deleted, the parent will be also deleted
        if subitem.text(0).startswith("frame") and ((item.childCount() == 1) or (item.childCount() == len(selected))):
            text = item.text(0)
            for action in actions:
                if action.text().endswith(text):
                    action.setEnabled(True)
                    item.takeChild(item.indexOfChild(subitem))
                    self.file_list.removeItemWidget(subitem, 0)
                    self.file_list.takeTopLevelItem(self.file_list.indexOfTopLevelItem(item))
                    self.file_list.removeItemWidget(item, 0)
                    del self.file_list.path_dict[text]
        else:
            for sel in selected:
                # tree_list elements get only reenabled in menu if the whole sheet is removed from file_list
                if sel.text(0).startswith("frame"):
                    continue

                for action in actions:
                    if action.text() == self.file_list.path_dict[sel.text(0)]:
                        action.setEnabled(True)

    def disable_files_in_menu_and_open(self, path):
        for action in self.main_window.recent_files_submenu.actions():
            if action.text() == path:
                action.setDisabled(True)
        try:
            self.file_list.add_file(path)
        except ValueError as e:
            QtWidgets.QMessageBox.information(self, "Information", str(e), QtWidgets.QMessageBox.Ok)
            return

    def open_file_dialog(self):
        filenames = QtWidgets.QFileDialog.getOpenFileNames(self, "Open dataset", self.most_recent_path)[0]
        for path in filenames:
            if path:
                self.disable_files_in_menu_and_open(path)
                self.main_window.update_submenu_recent_files()

    def calculationcallback(self, func, settings):
        def enable_screenshot_button_on_success(self, was_successful):
            if was_successful():
                self.file_list._enable_screenshot_button()

        self.thread = CalculationThread(self, func, settings)
        self.thread.finished.connect(
            functools.partial(
                self.control.update,
                was_successful=lambda: not self.thread.exited_with_errors,
            )
        )
        self.thread.finished.connect(
            functools.partial(
                self.main_window.updatestatus,
                was_successful=lambda: not self.thread.exited_with_errors,
            )
        )
        self.thread.finished.connect(
            functools.partial(
                enable_screenshot_button_on_success,
                self,
                was_successful=lambda: not self.thread.exited_with_errors,
            )
        )
        self.thread.start()
        self.progress_dialog.exec_()

    def calculate(self, file_frame_dict=None):
        if not file_frame_dict:
            file_frame_dict = self.file_list.get_selection()
        dia = CalculationSettingsDialog(self, file_frame_dict)
        settings, ok = dia.calculation_settings()

        if ok:
            self.control.calculationcallback = self.calculationcallback
            self.control.calculate(settings)
            self.last_shown_filename_with_frame = (
                list(file_frame_dict.keys())[-1],
                list(file_frame_dict.values())[-1][-1],
            )


class TreeList(QtWidgets.QTreeWidget):

    def __init__(self, parent, data={}):
        QtWidgets.QTreeWidget.__init__(self, parent)

        self.control = parent.control
        self.setColumnCount(1)
        self.path_dict = {}

        for root, sib in data.items():
            self.append_item(root, sib)

        self.setHeaderHidden(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setMouseTracking(True)
        self.itemSelectionChanged.connect(self.selection_changed)

    def append_item(self, root, sib):
        item = QtWidgets.QTreeWidgetItem(self)
        item.setText(0, root)
        if sib:
            for s in sib:
                tmp = QtWidgets.QTreeWidgetItem(item)
                tmp.setText(0, s)
                item.addChild(tmp)
        self.addTopLevelItem(item)

    def acceptable_drop_urls(self, e):
        if e.mimeData().hasUrls():
            for url in e.mimeData().urls():
                if url.scheme() == "file" and os.path.isfile(url.path()):
                    yield url

    def dragEnterEvent(self, e):
        if any(self.acceptable_drop_urls(e)):
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        if any(self.acceptable_drop_urls(e)):
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        if any(self.acceptable_drop_urls(e)):
            e.accept()
        else:
            e.ignore()
        # actually add the files
        for url in self.acceptable_drop_urls(e):
            self.add_file(url.path())

    def selection_changed(self):
        frames_selected = 0
        for item in self.selectedItems():
            # items representing the whole dataset
            if not item.data(0, 0).startswith("frame"):
                # select the children of the selected dataset
                for i in range(item.childCount()):
                    c = item.child(i)
                    c.setSelected(True)
                    frames_selected += 1
            else:
                frames_selected += 1
        any_frame_selected = len(self.selectedItems()) > 0
        self.parent().delete_button.setEnabled(any_frame_selected)
        self.parent().calculate_button.setEnabled(any_frame_selected)
        only_one_frame_available = self.topLevelItemCount() == 1 and self.topLevelItem(0).childCount() == 1
        self.parent().show_button.setEnabled(
            frames_selected == 1 or (not any_frame_selected and only_one_frame_available)
        )
        parent = self.parent()
        while parent.parent():
            parent = parent.parent()
        main_window = parent
        image_video_tab = main_window.image_video_dock.image_video_tab
        image_video_tab.mass_screenshot_button.setEnabled(any_frame_selected)
        image_video_tab.video_button.setEnabled(any_frame_selected)

    def files_changed(self):
        any_files_available = self.topLevelItemCount() > 0
        self.parent().select_all_button.setEnabled(any_files_available)
        self.parent().select_nth_button.setEnabled(any_files_available)
        self.parent().file_button.setDefault(not any_files_available)
        self.selection_changed()

    def get_selection(self):
        sel = {}
        for item in self.selectedItems():
            content = str(item.data(0, 0))
            if not content.startswith("frame"):
                sel[self.path_dict[content]] = [-1]
            else:
                parent_content = str(item.parent().data(0, 0))
                if self.path_dict[parent_content] in sel:
                    if not sel[self.path_dict[parent_content]][0] == -1:
                        sel[self.path_dict[parent_content]].append(int(content[6:]) - 1)
                else:
                    sel[self.path_dict[parent_content]] = [int(content[6:]) - 1]
        return sel

    def remove_selected_files(self):
        # delete childs
        del_files = self.selectedItems()
        for i, item in enumerate(del_files):
            content = str(item.data(0, 0))
            if content.startswith("frame"):
                item.parent().takeChild(item.parent().indexOfChild(item))
                self.removeItemWidget(item, 0)
                del del_files[i]

        # delete top level items
        for item in self.selectedItems():
            content = str(item.data(0, 0))
            if not content.startswith("frame"):
                self.takeTopLevelItem(self.indexOfTopLevelItem(item))
                self.removeItemWidget(item, 0)
                del self.path_dict[content]

        self.files_changed()

    def add_file(self, path):
        bname = os.path.basename(path)
        f = file.File.open(path)
        n_frames = f.info.num_frames

        if bname not in self.path_dict.keys() and File.exists(path):
            self.path_dict[bname] = str(path)
            root = bname
            sib = ["frame {}".format(i) for i in range(1, n_frames + 1)]
            self.append_item(root, sib)

            if path not in config.recent_files:
                config.add_recent_file(path)
            else:
                index = config.recent_files.index(path)
                config.recent_files.pop(index)
                config.add_recent_file(path)

        self.files_changed()

    def show_selected_frame(self):
        # if there is just one file with one frame, select that frame automatically
        if self.topLevelItemCount() == 1 and self.topLevelItem(0).childCount() == 1:
            self.select_all()

        sel = self.get_selection()

        filename = list(sel.keys())[0]
        frame = sel[filename][0]

        if frame == -1:
            if file.File.open(filename).info.num_frames > 1:
                frame = -2
            else:
                frame = 0

        f = file.File.open(filename)
        if filename not in self.parent().guessed_volumes_for and f.info.volume_guessed:
            msgBox = QtWidgets.QMessageBox(self)
            msgBox.setWindowTitle("Missing cell shape description")
            msgBox.setTextFormat(QtCore.Qt.RichText)
            msgBox.setText(
                "This file does not contain information about the "
                "cell shape. Should pyMolDyn use an orthorhombic "
                "shape?<br />"
                "You can find information about the different cell "
                "shapes shapes in the "
                '<a href="https://pgi-jcns.fz-juelich.de/portal/pages/pymoldyn-doc.html#cell-shape-description">pyMolDyn documentation</a>'
            )
            msgBox.addButton(QtWidgets.QMessageBox.Yes)
            msgBox.addButton(QtWidgets.QMessageBox.No)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Yes)
            msgBox.setEscapeButton(QtWidgets.QMessageBox.No)
            response = msgBox.exec_()
            if response == QtWidgets.QMessageBox.No:
                return
            self.parent().guessed_volumes_for.add(filename)

        self.control.visualize(filename, frame)

        widget = self
        while not hasattr(widget, "updatestatus"):
            widget = widget.parent()
        widget.updatestatus()

        # update GL scene
        for widget in QtWidgets.QApplication.topLevelWidgets():
            for gl_widget in widget.findChildren(GLWidget):
                gl_widget.update_needed = True
                QtWidgets.QApplication.postEvent(gl_widget, UpdateGLEvent())

        self._enable_screenshot_button()

        return filename, frame

    def _enable_screenshot_button(self, enable=True):
        parent = self.parent()
        while parent.parent():
            parent = parent.parent()
        main_window = parent
        image_video_tab = main_window.image_video_dock.image_video_tab
        image_video_tab.screenshot_button.setEnabled(enable)

    def select_all(self):
        self.select_nth(1)

    def select_nth(self, n):
        assert int(n) == n and n > 0
        for file_index in range(self.topLevelItemCount()):
            item = self.topLevelItem(file_index)
            if n == 1:
                item.setSelected(True)
            else:
                item.setSelected(False)
                item.setExpanded(True)
            # select every nth child and unselect the others
            for frame_index in range(item.childCount()):
                c = item.child(frame_index)
                c.setSelected(frame_index % n == 0)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from PySide6 import QtCore, QtWidgets
import gr
from qtgr import GRWidget
import csv

from ..util.logger import Logger
import sys
from ..statistics.pdf import PDF, Kernels
import numpy as np
import os


logger = Logger("gui.pdf_widget")
logger.setstream("default", sys.stdout, Logger.DEBUG)


class GrPlotWidget(GRWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.xvalues = None
        self.yvalues = None
        self.title = None
        self.datapoints = None

    def quit(self):
        gr.emergencyclosegks()
        self.close()

    def setdata(self, xvalues, yvalues, title, datapoints=None):
        self.xvalues = xvalues
        self.yvalues = yvalues
        self.title = title
        self.datapoints = datapoints

    def draw(self, wsviewport=None):
        if self.xvalues is not None:
            rangex = (self.xvalues.min(), self.xvalues.max())
        else:
            rangex = (0, 10)
        if self.yvalues is not None:
            rangey = gr.adjustrange(self.yvalues.min(), self.yvalues.max())
        else:
            rangey = (0, 4)

        if wsviewport is None:
            gr.setwsviewport(0, self.mwidth, 0, self.mheight)
        else:
            gr.setwsviewport(*wsviewport)
        gr.setwswindow(0, self.sizex, 0, self.sizey)
        gr.setviewport(0.075 * self.sizex, 0.95 * self.sizex, 0.075 * self.sizey, 0.95 * self.sizey)
        gr.setwindow(rangex[0], rangex[1], rangey[0], rangey[1])
        gr.setcharheight(0.012)

        gr.setfillintstyle(1)
        gr.setfillcolorind(0)
        gr.fillrect(rangex[0], rangex[1], rangey[0], rangey[1])

        if self.xvalues is not None and self.yvalues is not None:
            gr.setlinecolorind(2)
            gr.polyline(self.xvalues, self.yvalues)
        else:
            gr.text(0.4 * self.sizex, 0.5 * self.sizey, "no elements selected")

        gr.setlinecolorind(1)
        gr.axes(0.2, 0.2, rangex[0], rangey[0], 5, 5, 0.0075)
        gr.axes(0.2, 0.2, rangex[1], rangey[1], -5, -5, -0.0075)

        if self.title is not None:
            gr.text(0.8 * self.sizex, 0.9 * self.sizey, self.title)


class PDFWidget(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__(parent)

        self.control = parent.control
        self.results = None
        self.pdf = None

        self.init_gui()

    def init_gui(self):
        vbox = QtWidgets.QVBoxLayout()
        grid = QtWidgets.QGridLayout()
        self.gr_widget = GrPlotWidget()

        self.datasetlabel = QtWidgets.QLabel("No data loaded.", self)
        self.datasetlabel.setAlignment(QtCore.Qt.AlignHCenter)

        elembox = QtWidgets.QHBoxLayout()
        elembox.addWidget(QtWidgets.QLabel("Elements:", self), 0)
        self.elem1 = QtWidgets.QComboBox(self)
        self.elem1.setMinimumWidth(170)
        elembox.addWidget(self.elem1, 1)
        self.elem2 = QtWidgets.QComboBox(self)
        self.elem2.setMinimumWidth(170)
        elembox.addWidget(self.elem2, 1)
        grid.addLayout(elembox, 0, 0)

        rangebox = QtWidgets.QHBoxLayout()
        rangebox.addWidget(QtWidgets.QLabel("Plot range:", self))
        self.range1 = QtWidgets.QLineEdit("0", self)
        self.range1.setMinimumWidth(30)
        rangebox.addWidget(self.range1)
        rangebox.addWidget(QtWidgets.QLabel("-", self))
        self.range2 = QtWidgets.QLineEdit("8", self)
        self.range2.setMinimumWidth(30)
        rangebox.addWidget(self.range2)
        grid.addLayout(rangebox, 1, 0)

        cutoffbox = QtWidgets.QHBoxLayout()
        cutoffbox.addWidget(QtWidgets.QLabel("Kernel:", self))
        self.kernels = {
            "Gaussian": Kernels.gauss,
            "Epanechnikov": Kernels.epanechnikov,
            "Compact": Kernels.compact,
            "Triangular": Kernels.triang,
            "Box": Kernels.quad,
            "Right Box": Kernels.posquad,
            "Left Box": Kernels.negquad,
        }
        self.kernel = QtWidgets.QComboBox(self)
        self.kernel.setMinimumWidth(130)
        self.kernel.addItems(
            [
                "Gaussian",
                "Epanechnikov",
                "Compact",
                "Triangular",
                "Box",
                "Right Box",
                "Left Box",
            ]
        )
        cutoffbox.addWidget(self.kernel)
        cutoffbox.addWidget(QtWidgets.QLabel("Cutoff:", self))
        self.cutoff = QtWidgets.QLineEdit("12", self)
        self.cutoff.setMinimumWidth(30)
        cutoffbox.addWidget(self.cutoff)
        cutoffbox.addWidget(QtWidgets.QLabel("Bandwidth:", self))
        self.bandwidth = QtWidgets.QLineEdit("", self)
        self.bandwidth.setMinimumWidth(30)
        cutoffbox.addWidget(self.bandwidth)
        grid.addLayout(cutoffbox, 2, 0)

        buttonbox = QtWidgets.QHBoxLayout()

        self.plotbutton = QtWidgets.QPushButton("Plot", self)
        buttonbox.addWidget(self.plotbutton)
        self.plotbutton.clicked.connect(self.draw)

        self.export_image_button = QtWidgets.QPushButton("Save Image", self)
        buttonbox.addWidget(self.export_image_button)
        self.export_image_button.clicked.connect(self.export_image)

        self.export_data_button = QtWidgets.QPushButton("Export Data", self)
        buttonbox.addWidget(self.export_data_button)
        self.export_data_button.clicked.connect(self.export_data)
        grid.addLayout(buttonbox, 3, 0)

        vbox.addWidget(self.gr_widget, stretch=1)
        vbox.addWidget(self.datasetlabel, stretch=0)
        vbox.addLayout(grid)
        self.setLayout(vbox)
        self.show()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Return or event.key() == QtCore.Qt.Key_Enter:
            self.update()

    def draw(self, now=False, wsviewport=None):
        xvalues = None
        yvalues = None
        title = None
        datapoints = None
        if self.pdf is None:
            self.refresh()
        if self.pdf is not None:
            elem1 = str(self.elem1.currentText())
            if elem1 == "cavity domain centers":
                elem1 = "cav"
            elem2 = str(self.elem2.currentText())
            if elem2 == "cavity domain centers":
                elem2 = "cav"
            range1 = float(str(self.range1.text()))
            range2 = float(str(self.range2.text()))
            cutoff = str(self.cutoff.text())
            if len(cutoff) > 0 and float(cutoff) > 0:
                cutoff = float(cutoff)
            else:
                cutoff = None
            try:
                bandwidth = float(str(self.bandwidth.text()))
                if bandwidth < 0:
                    bandwidth = 0
                    self.bandwidth.setText("0")
            except ValueError:
                bandwidth = None
                self.bandwidth.setText("")
            kernel = self.kernels.get(self.kernel.currentText(), None)
            f = self.pdf.pdf(elem1, elem2, cutoff=cutoff, h=bandwidth, kernel=kernel)
            if f is not None:
                if callable(f):
                    xvalues = np.linspace(range1, range2, 400)
                    yvalues = f(xvalues)
                    title = "{} - {}".format(elem1, elem2)
                    datapoints = f.f.x
                else:
                    peaks = f[np.logical_and(range1 < f, f < range2)]
                    if len(peaks) > 2:
                        xvalues = np.zeros(len(peaks) * 3)
                        xvalues[0::3] = peaks
                        xvalues[1::3] = peaks
                        xvalues[2::3] = peaks
                        yvalues = np.zeros(len(peaks) * 3)
                        yvalues[1::3] = 1
                        datapoints = peaks

        self.gr_widget.setdata(xvalues, yvalues, title, datapoints)
        self.gr_widget.update()
        if now:
            self.gr_widget.draw(wsviewport=wsviewport)

    def export_image(self):
        extensions = (
            ".pdf",
            ".png",
            ".bmp",
            ".jpg",
            ".jpeg",
            ".png",
            ".tiff",
            ".fig",
            ".svg",
            ".wmf",
            ".eps",
            ".ps",
        )
        qtext = "*" + " *".join(extensions)
        filepath = QtWidgets.QFileDialog.getSaveFileName(self, "Save Image", ".", "Image Files ({})".format(qtext))[0]
        if len(filepath) == 0:
            return

        if filepath.endswith(".eps") or filepath.endswith(".ps"):
            gr.beginprintext(filepath, "Color", "A4", "Landscape")
            self.draw(now=True, wsviewport=(0, 0.297 * 0.9, 0, 0.21 * 0.95))
        else:
            gr.beginprint(filepath)
            self.draw(now=True)
        gr.endprint()

    def export_data(self):
        qtext = " *.csv"
        filepath = QtWidgets.QFileDialog.getSaveFileName(self, "Save Data", ".", "CSV Files ({})".format(qtext))[0]
        if len(filepath) == 0:
            return
        self.update()
        xvalues = self.gr_widget.xvalues
        yvalues = self.gr_widget.yvalues
        if xvalues is None or yvalues is None:
            return
        with open(filepath, "w") as csvfile:  # TODO same check as with histogramm (bytes csv)
            csvwriter = csv.writer(csvfile)
            for x, y in zip(xvalues, yvalues):
                csvwriter.writerow([x, y])

    def refresh(self):
        results = self.control.results
        if results is not None:
            results = results[-1][-1]
            if self.results != results or self.pdf is None:
                self.results = results
                self.pdf = PDF(results)
                e = np.unique(results.atoms.elements).tolist()
                for i in range(len(e)):
                    e[i] = e[i].decode("utf-8")
                if results.domains is not None and len(results.domains.centers) > 0 and "cav" not in e:
                    e.append("cavity domain centers")
                self.elem1.clear()
                self.elem1.addItems(e)
                self.elem2.clear()
                self.elem2.addItems(e)
                self.gr_widget.setdata(None, None, None)
                self.gr_widget.update()
            self.datasetlabel.setText(str(results))
        else:
            self.datasetlabel.setText("")
            self.elem1.clear()
            self.elem2.clear()

    def activate(self):
        self.refresh()

    def updatestatus(self):
        self.refresh()



# -*- coding: utf-8 -*-

import re
from PySide6 import QtCore, QtGui, QtWidgets


class ObjectSelectWidget(QtWidgets.QWidget):
    state_changed = QtCore.Signal(bool)
    selection_indices_changed = QtCore.Signal(object)

    def __init__(self, text, parent, add_index_selector=True, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        self._has_index_selector = add_index_selector

        # Widgets
        self._activation_checkbox = QtWidgets.QCheckBox(text, self)
        if self._has_index_selector:
            self._selection_checkbox = QtWidgets.QCheckBox("only selected indices", self)
            self._index_selection = IndexSelectLineEdit(self)

        # Layout
        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setColumnMinimumWidth(0, 8)
        layout.addWidget(self._activation_checkbox, 0, 0, 1, 3)
        if self._has_index_selector:
            layout.addWidget(self._selection_checkbox, 1, 1)
            layout.addWidget(self._index_selection, 1, 2)
        self.setLayout(layout)

        # Signals
        self._activation_checkbox.stateChanged.connect(self._state_changed)
        if self._has_index_selector:
            self._selection_checkbox.stateChanged.connect(self._selection_toggled)
            self._index_selection.indices_changed.connect(self._selection_indices_changed)

        # State setup
        self._state_changed(check_state=False)

    @property
    def indices(self):
        if self._has_index_selector and self._selection_checkbox.isChecked():
            return self._index_selection.indices
        else:
            return None

    @indices.setter
    def indices(self, indices):
        if self._has_index_selector:
            self._index_selection.indices = indices

    def add_indices(self, indices):
        if self._has_index_selector:
            self._index_selection.add_indices(indices)

    def isChecked(self):
        return self._activation_checkbox.isChecked()

    def setChecked(self, checkState):
        return self._activation_checkbox.setChecked(checkState)

    def selection_checkbox_is_checked(self):
        return self._selection_checkbox.isChecked()

    def selection_checkbox_set_checked(self, check_state):
        return self._selection_checkbox.setChecked(check_state)

    def _state_changed(self, check_state):
        if self._has_index_selector:
            self._selection_checkbox.setEnabled(check_state)
            self._index_selection.setEnabled(check_state)
        self.state_changed.emit(check_state)

    def _selection_toggled(self, check_state):
        if self._index_selection.indices is not None:
            self.selection_indices_changed.emit(self.indices)

    def _selection_indices_changed(self, indices):
        if self._selection_checkbox.isChecked():
            self.selection_indices_changed.emit(indices)


class IndexSelectLineEdit(QtWidgets.QLineEdit):
    indices_changed = QtCore.Signal(object)
    TIMER_INTERVAL = 1000

    class Validator(QtGui.QValidator):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def validate(self, input, pos):
            return (
                (QtGui.QValidator.Acceptable if re.match(r"^[0-9,\s-]*$", input) else QtGui.QValidator.Invalid),
                input,
                pos,
            )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._indices = None
        self._next_indices = None
        self.setValidator(IndexSelectLineEdit.Validator())
        self.indices_emit_timer = QtCore.QTimer(self)
        self.indices_emit_timer.setInterval(self.TIMER_INTERVAL)
        self.indices_emit_timer.setSingleShot(True)
        self.indices_emit_timer.timeout.connect(self._time_out)
        self.textEdited.connect(self._text_edited)

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, indices):
        def indices2string(indices):
            def range2string(range):
                if range is None:
                    return ""
                elif len(range) == 1:
                    start = range[0]
                    end = start
                else:
                    start, end = range
                if start == end:
                    return str(start + 1)
                else:
                    return "{:d}-{:d}".format(start + 1, end + 1)

            indices = list(set(indices))
            indices.sort()
            ranges = []
            if len(indices) == 0:
                current_range = None
            else:
                current_range = [indices[0]]
                previous_index = indices[0]
                for current_index in indices[1:]:
                    if current_index > previous_index + 1:
                        current_range.append(previous_index)
                        ranges.append(current_range)
                        current_range = [current_index]
                    previous_index = current_index
                current_range.append(previous_index)
            ranges.append(current_range)
            indices_string = ", ".join(map(range2string, ranges))
            return indices_string

        self.setText(indices2string(indices))
        self._indices = tuple(indices)
        if self.isEnabled():
            self.indices_changed.emit(self._indices)

    def add_indices(self, indices):
        if self._indices is not None:
            self.indices = set(self._indices) | set(indices)
        else:
            self.indices = indices

    def _text_edited(self):
        if self.indices_emit_timer.isActive():
            self.indices_emit_timer.stop()
        indices_from_line_edit = self._get_indices_from_line_edit()
        if indices_from_line_edit is not None:
            if len(indices_from_line_edit) > 0:
                self._next_indices = indices_from_line_edit
            else:
                self._next_indices = None
            self.indices_emit_timer.start()

    def _time_out(self):
        if self._next_indices != self._indices:
            self._indices = self._next_indices
            self.indices_changed.emit(self._indices)

    def _get_indices_from_line_edit(self):
        input_text = self.text().replace(" ", "")
        is_input_valid = re.match(r"^(\d+(-\d+)?(,\d+(-\d+)?)*)?$", input_text)
        if is_input_valid:
            parts = input_text.split(",") if input_text != "" else []
            indices = []
            indices.extend((int(comp) - 1 for comp in parts if "-" not in comp))  # single indices
            for comp in parts:
                if "-" in comp:  # index ranges
                    start, end = map(int, comp.split("-"))
                    indices.extend(range(start - 1, end))
            indices = list(set(indices))  # Eliminate duplicate entries
            indices.sort()
            return tuple(indices)
        else:
            return None
